import warnings
import bincfg
from collections import Counter
from .cfg_edge import get_edge_type, EdgeType, CFGEdge
from ..utils import get_address, eq_obj, hash_obj
from ..utils.type_utils import *


if IN_PYTHON_TYPING_VERSION:
    CFGBasicBlockPickledState = Tuple[int, Tuple[Tuple[int, int, EdgeType], ...], Tuple[Tuple[int, int, EdgeType], ...], list[str], list[int], dict]
    """The pickled state of a CFGBasicBlock"""


class CFGBasicBlock:
    """A single basic block in a ``CFG``.

    Can be initialized empty, or with attributes. Assumes its memory address is always unique within a cfg.

    NOTE: these objects should not be pickled/copy.deepcopy()-ed by themselves, only as a part of a cfg

    Parameters
    ----------
    parent_function: `Optional[CFGFunction]`
        the ``CFGFunction`` this basic block belongs to
    address: `Optional[Union[int, str, Addressable]]`
        the memory address of this ``CFGBasicBlock``. Should be unique to the ``CFG`` that contains it. If None, but
        `asm_memory_addresses` is passed, this will be set to the first value in `asm_memory_addresses`
    edges_in: `Optional[Iterable[CFGEdge]]` 
        an iterable of incoming CFGEdge objects
    edges_out: `Optional[Iterable[CFGEdge]]`
        an iterable of outgoing CFGEdge objects
    asm_lines: `Optional[Iterable[str]]`
        an iterable of string assembly lines present at this basic block
    asm_memory_addresses: `Optional[Iterable[Union[str, int, Addressable]]]`
        an iterable of string or integer memory addresses, one for each assembly line (will be converted into integer 
        memory addresses). If this was passed, but `address` was not, then `address` will be set to the first value in
        `asm_memory_addresses`
    metadata: `Optional[Dict]`
        optional dictionary of metadata to associate with this basic block
    """

    parent_function: 'Union[bincfg.CFGFunction, None]'
    """The parent function of this basic block. Will be None if not set yet"""
    address: 'int'
    """The integer memory address of this basic block. Will be -1 if not set yet"""
    edges_in: 'set[CFGEdge]'
    """The set of incomming ``CFGEdge``'s to this basic block"""
    edges_out: 'set[CFGEdge]'
    """The set of outgoing ``CFGEdge``'s from this basic block"""
    asm_lines: 'list[str]'
    """List of string assembly lines at this basic block"""
    asm_memory_addresses: 'list[int]'
    """List of integer memory addresses for all assembly lines at this basic block. Will be empty list if not set yet"""
    metadata: 'dict'
    """Dictionary of extra metadata to associate with this basic block"""


    def __init__(self, parent_function: 'Optional[bincfg.CFGFunction]' = None, address: 'Optional[AddressLike]' = None, 
                 edges_in: 'Optional[Iterable[CFGEdge]]' = None, edges_out: 'Optional[Iterable[CFGEdge]]' = None, 
                 asm_lines: 'Optional[Iterable[str]]' = None, asm_memory_addresses: 'Optional[Iterable[AddressLike]]' = None, 
                 metadata: 'Optional[dict]' = None):
        # Get the memory addresses all figured out
        self.asm_memory_addresses = [] if asm_memory_addresses is None else [get_address(addr) for addr in asm_memory_addresses]
        self.asm_memory_addresses: 'list[int]' = [] if all(a < 0 for a in self.asm_memory_addresses) else self.asm_memory_addresses

        if address is None and len(self.asm_memory_addresses) > 0:
            self.address = self.asm_memory_addresses[0]
        else:
            self.address = get_address(address) if address is not None else -1

        # Set the rest of the params
        self.parent_function = parent_function
        self.edges_in = set() if edges_in is None else set(edges_in)
        self.edges_out = set() if edges_out is None else set(edges_out)
        self.asm_lines = [] if asm_lines is None else list(asm_lines)
        self.metadata = {} if metadata is None else metadata
    
    @property
    def num_edges(self) -> 'int':
        """The number of edges out in this basic block"""
        return self.num_edges_out
    
    @property
    def num_edges_out(self) -> 'int':
        """The number of outgoing edges in this basic block"""
        return len(self.edges_out)

    @property
    def num_edges_in(self) -> 'int':
        """The number of incoming edges in this basic block"""
        return len(self.edges_in)
    
    @property
    def num_asm_lines(self) -> 'int':
        """The number of assembly lines/tokens in this basic block"""
        return len(self.asm_lines)
    
    @property
    def asm_counts(self) -> 'Mapping[str, int]':
        """A ``collections.Counter`` of all unique assembly lines/tokens and their counts in this basic block"""
        return Counter(l for l in self.asm_lines)
    
    @property
    def is_function_entry(self) -> 'bool':
        """True if this block is a function entry block, False otherwise
        
        Specifically, returns True if this block's address matches its parent function's address. If this block has
        no parent, False is returned.
        """
        return self.parent_function is not None and self.address == self.parent_function.address
    
    @property
    def is_function_call(self) -> 'bool':
        """True if this block is a function call, False otherwise
        
        Checks if this block has one or more outgoing function call edges
        """
        return any(e.edge_type is EdgeType.FUNCTION_CALL for e in self.edges_out)

    @property
    def is_function_jump(self) -> 'bool':
        """True if this block is a function jump, False otherwise
        
        Checks if this block has a 'jump' instruction to a basic block in a different function. Specifically, checks if
        this block has an outgoing EdgeType.NORMAL edge to a basic block who's parent_function has an address different
        than this basic block's parent_function's address.
        """
        return self.parent_function is not None and \
            any((e.edge_type is EdgeType.NORMAL and e.to_block.parent_function is not None 
                 and e.to_block.parent_function.address != self.parent_function.address) for e in self.edges_out)
    
    @property
    def is_multi_function_call(self) -> 'bool':
        """True if this block is a multi-function call, False otherwise
        
        IE: this block has either two or more function call edges out
        """
        return len(self.get_sorted_edges(edge_types='function_call', direction='out')[0]) >= 2
    
    @property
    def all_edges(self) -> 'set[CFGEdge]':
        """Returns a set of all edges in this basic block"""
        return self.edges_in.union(self.edges_out)
    
    def remove_edge(self, edge: 'CFGEdge') -> None:
        """Removes the given edge from this block's edges (both incoming and outgoing)
        
        Args:
            edge (CFGEdge): the CFGEdge to remove
        
        Raises:
            ValueError: if the edge doesn't exist in the incomming/outgoing edges
        """
        if not isinstance(edge, CFGEdge):
            raise TypeError("edge must be a CFGEdge, not %s" % repr(type(edge).__name__))
        if edge not in self.edges_in and edge not in self.edges_out:
            raise ValueError("%s does not exist in this block's (%s) edges in or out!" % 
                             (edge, ("0x%x" % self.address) if self.address is not None else "NoAddr"))
        
        if edge in self.edges_in:
            self.edges_in.remove(edge)
        if edge in self.edges_out:
            self.edges_out.remove(edge)

    def has_edge(self, address: 'AddressLike', edge_types: 'Optional[Union[str, EdgeType, Iterable[Union[str, EdgeType]]]]' = None, 
                 direction: "Optional[Literal['in', 'out']]" = None) -> 'bool':
        """Checks if this block has an edge from/to the given address

        Args:
            address (AddressLike): a string/integer memory address, or an addressable object (EG: ``CFGBasicBlock``/``CFGFunction``). 
            edge_types (Optional[Union[str, EdgeType, Iterable[Union[str, EdgeType]]]]): either an edge type or an
                iterable of edge types. Only edges with one of these types will be considered. If None, then all edge 
                types will be considered
            direction (Optional[Literal['in', 'out']]): the direction to check (strings 'in' or 'out), or None to check both

        Returns:
            bool: True if this block has an edge from/to the given address, False otherwise
        """
        addr, edge_types, directions = get_address(address), _get_edge_types(edge_types, as_set=True), self._get_directions(direction)

        # Check for that edge
        for edge_set in directions:
            for edge in edge_set:
                # Check both that it is an allowable edge_type, and that the edge from/to address matches the given address
                if edge.edge_type in edge_types and \
                    ((edge_set is self.edges_in and edge.from_block.address == addr) or (edge_set is self.edges_out and edge.to_block.address == addr)):
                    return True

        return False
    
    def calls(self, address: 'AddressLike'):
        """Checks if this block calls the given address

        IE: checks if this block has an outgoing `function_call` edge to the given address

        Args:
            address (AddressLike): a string/integer memory address, or an addressable object (EG: ``CFGBasicBlock``/``CFGFunction``)

        Returns:
            bool: True if this block calls the given address, False otherwise
        """
        return self.has_edge(address=address, edge_types='function_call', direction='to')
    
    def get_sorted_edges(self, edge_types: 'Optional[Union[str, EdgeType, Iterable[Union[str, EdgeType]]]]' = None, 
                         direction: 'Optional[Union[Literal["out", "in"], Iterable[Literal["out", "in"]]]]' = None, 
                         as_sets: bool = False) -> 'Union[Tuple[list[CFGEdge], ...], Tuple[set[CFGEdge], ...]]':
        """Returns a tuple of sorted lists of edges (sorted by address of the "other" block) of each type/direction in this block
        
        Will return edge lists ordered first by edge type (their order of appearance in the cfg_edge.EdgeType enum),
        then by direction ('in', then 'out'). Unless, if `edge_types` is passed, then only those edge types will be
        returned and the edge lists will be returned in the order of the edge types in `edge_types`, then by direction
        ('in', then 'out').

        For example, with `edge_types=None` and `direction=None`, this would return the 4-tuple of:
        (normal_edges_in, normal_edges_out, function_call_edges_in, function_call_edges_out)
        Where each element is a list of CFGEdge objects.

        Args:
            edge_types (Optional[Union[str, EdgeType, Iterable[Union[str, EdgeType]]]]): either an edge type or an
                iterable of edge types. Only edges with one of these types will be returned. If not None, then the edge 
                lists will be returned sorted based on the order of the edge types listed here, then by direction
            direction (Optional[Union[Literal["out", "in"], Iterable[Literal["out", "in"]]]): the direction to get. Can 
                be the strings 'in' or 'out', or None to get both
            as_sets (bool): if True, then this will return unordered sets of edges instead of sorted lists. This may
                save a ~tiny~ bit of time in the long run, but will hinder deterministic behavior of this method.

        Returns:
            Union[Tuple[List[CFGEdge], ...], Tuple[Set[CFGEdge], ...]]: a tuple of lists/sets of CFGEdge's
        """
        edge_types, directions = _get_edge_types(edge_types, as_set=False), self._get_directions(direction)

        # Iterate through all edges, adding them to their appropriate sets
        ret_sets = [set() for d in directions for et in edge_types]
        for i, edge_set in enumerate(directions):
            for edge in edge_set:
                if edge.edge_type in edge_types:
                    ret_sets[edge_types.index(edge.edge_type) * len(directions) + i].add(edge)
        
        # Check if we need to sort the edges. If so, sort based on the address of the "other" block
        return ret_sets if as_sets else [list(sorted(s, key=lambda edge: 
                (edge.from_block.address if directions[i % len(directions)] is self.edges_in else edge.to_block.address))) 
            for i, s in enumerate(ret_sets)]

    def _get_directions(self, direction: 'Literal["in", "out", "to", "from"]') -> 'list[list[CFGEdge]]':
        """Gets the directions

        Args:
            direction (Union[str, None]): the direction to get (strings 'in'/'from' or 'to'/'out'), or None to get both 
                (in order ['in', 'out'])

        Raises:
            ValueError: on an unknown direction

        Returns:
            List[List[CFGEdge]]: a list of edges in/out based on direction
        """
        if direction is None:
            return [self.edges_in, self.edges_out]
        elif isinstance(direction, str):
            if direction not in ['from', 'to', 'in', 'out']:
                raise ValueError("Unknown `direction`: %s" % repr(direction))
            return [self.edges_in] if direction in ['from', 'in'] else [self.edges_out]
        else:
            return [self._get_directions(v)[0] for v in direction]
        
    def __str__(self) -> 'str':
        valid_pf = isinstance(self.parent_function, bincfg.CFGFunction)
        asm = '\n'.join([("\t0x%s: %s" % ('%08x' % addr, line)) for addr, line in zip(self.asm_memory_addresses, self.asm_lines)]) \
            if len(self.asm_memory_addresses) > 0 else '\n'.join([("\t%s" % line) for line in self.asm_lines])
        func_name = '' if not valid_pf or self.parent_function.name is None else self.parent_function.name

        func_str = ('in function \"%s\"' % func_name) if valid_pf else 'with NO PARENT'
        addr_str = ('0x%08x' % self.address) if self.address is not None else 'NO_ADDRESS'

        ret = "CFGBasicBlock %s at %s with %d edges out, %d edges in, with %d lines of assembly:\n%s\nEdges Out: %s\nEdges In: %s\n" \
            % (func_str, addr_str, self.num_edges_out, self.num_edges_in, len(self.asm_lines), asm, self.edges_out, self.edges_in)
        
        return (ret + "Metadata: %s" % self.metadata) if len(self.metadata) > 0 else ret
    
    def __repr__(self) -> 'str':
        return str(self)
    
    def __eq__(self, other: 'Any') -> 'bool':
        return isinstance(other, CFGBasicBlock) and all(eq_obj(self, other, selector=s) for s in \
            ['address', 'edges_in', 'edges_out', 'metadata', 'asm_memory_addresses', 'asm_lines'])
    
    def __hash__(self) -> 'int':
        return hash_obj([self.address, self.edges_in, self.edges_out, self.metadata, self.asm_lines, self.asm_memory_addresses], return_int=True)
    
    def __getstate__(self) -> 'CFGBasicBlockPickledState':
        """Print a warning about pickling singleton basic block objects"""
        warnings.warn("Attempting to pickle a singleton basic block object! This will mess up edges unless you know what you're doing!")
        return self._get_pickle_state()
    
    def __setstate__(self, state: 'CFGBasicBlockPickledState') -> 'None':
        """Print a warning about pickling singleton basic block objects"""
        warnings.warn("Attempting to unpickle a singleton basic block object! This will mess up edges unless you know what you're doing!")
        self._set_pickle_state(state)
    
    def _get_pickle_state(self) -> 'CFGBasicBlockPickledState':
        """Returns info of this CFGBasicBlock as a tuple"""
        edges_in = tuple((e.from_block.address, e.to_block.address, e.edge_type) for e in self.edges_in)
        edges_out = tuple((e.from_block.address, e.to_block.address, e.edge_type) for e in self.edges_out)
        return (self.address, edges_in, edges_out, self.asm_lines, self.asm_memory_addresses, self.metadata)
    
    def _set_pickle_state(self, state: 'CFGBasicBlockPickledState') -> 'CFGBasicBlock':
        """Set the pickled state, looking at parent function.parent_cfg for info for building edges"""
        self.address, self._temp_edges_in, self._temp_edges_out, self.asm_lines, self.asm_memory_addresses, self.metadata = state
        return self


def _get_edge_types(edge_types: 'Optional[Union[str, EdgeType, Iterable[Union[str, EdgeType]]]]' = None, as_set: bool = False):
    """Gets the edge types passed by the user
    
    Args:
        edge_types (Optional[Union[str, EdgeType, Iterable[Union[str, EdgeType]]]]): either an edge type or an 
            iterable of edge types. Only edges with one of these types will be considered. Defaults to None.
        as_set (bool): if True, will return the result as a set instead. Defaults to False.

    Raises:
        TypeError: on an unknown `edge_types`

    Returns:
        Union[List[EdgeType], Set[EdgeType]]: a list/set of EdgeType objects
    """
    ret_type = set if as_set else list

    if edge_types is None:
        return ret_type(e for e in EdgeType)
    else:
        # Attempt a single EdgeTypeLike first
        try:
            return ret_type([get_edge_type(edge_types)])
        except (ValueError, TypeError):
            try:
                return ret_type(get_edge_type(et) for et in edge_types)
            except Exception:
                raise TypeError("Could not get acceptable edge types from `edge_types` of type %s" % type(edge_types).__name__)



from .cfg import CFG
from .cfg_function import CFGFunction
from .cfg_basic_block import CFGBasicBlock
from .cfg_edge import CFGEdge, EdgeType
from .cfg_dataset import CFGDataset
from .mem_cfg import MemCFG
from .mem_cfg_dataset import MemCFGDataset



"""
A bunch of builtin normalization methods based on literature.

NOTE: some of these are slightly modified from their original papers either for code purposes, or because we are using
decompiled binaries instead of compiled assembly and thus lose out on some information (EG: symbol information
for jump instructions)
"""
from ..base_normalizer import TokenizationLevel, BaseNormalizer, LIBC_FUNCTION_NAMES
from .x86_tokenizer import X86_DEFAULT_TOKENIZER, X86_RE_SEGMENT_REG
from .x86_norm_funcs import *
from ..norm_utils import SPLIT_IMMEDIATE_TOKEN, scan_for_token, RE_SPACING
from ..norm_funcs import *
from ...utils import eq_obj, hash_obj


# Used to replace '{spacing}ptr' in a memory size expression
REPL_SPACE_PTR = re.compile(r'{space}ptr'.format(space=RE_SPACING), flags=re.IGNORECASE)


class X86BaseNormalizer(BaseNormalizer):
    """Base class for x86 normalizers.
    
    Performs an 'unnormalized' normalization, removing what is likely extraneous information, and providing a base class
    for other x86 normalization methods to inherit from.
    
    Parameters
    ----------
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """The default tokenization level used for this normalizer"""

    renormalizable = True
    """Whether or not this normalization method can be renormalized later by other normalization methods"""

    tokenizer = None
    """The tokenizer used for this normalizer"""

    token_sep = None
    """The separator string used for this normalizer. Will default to ' '"""

    tokenization_level: TokenizationLevel
    """The tokenization level to use for this normalizer"""

    def __init__(self, tokenizer=None, token_handlers=None, token_sep=None, tokenization_level=TokenizationLevel.AUTO, anonymize_tokens=False):
        # Add the handle_segment_address() function
        token_handlers = {} if token_handlers is None else token_handlers
        token_handlers.setdefault(Tokens.SEGMENT_ADDRESS, self.handle_segment_address)

        super().__init__(tokenizer=tokenizer if tokenizer is not None else X86_DEFAULT_TOKENIZER, token_handlers=token_handlers,
                         token_sep=token_sep if token_sep is not None else ' ',
                         tokenization_level=tokenization_level, anonymize_tokens=anonymize_tokens)
        self.register_opcode_handler(r'(?:far)?callq?', self.opcode_function_call)
        self.register_opcode_handler(r'j.*|farjmp', self.opcode_jump)
        self.register_opcode_handler(r'nop.*', x86_clean_nop)
    
    def opcode_function_call(self, state):
        """Handles function call opcodes, defaults to doing nothing"""
        return None
    
    def opcode_jump(self, state):
        """Handles jump opcodes, defaults to doing nothing"""
        return None
    
    def handle_all_symbols(self, state):
        """Handles all symbols. We use this to keep track of when memory expressions start/end"""
        # Check for memory address brackets
        if state.token_type in [Tokens.OPEN_BRACKET]:
            state.set(memory_start=len(state.line))
        elif state.token_type in [Tokens.CLOSE_BRACKET]:
            state.line.append((state.token_type, state.token, state.token))
            self.handle_memory_expression(state)
            state.set(memory_start=None)
            return None  # Don't append the token since we've already done that

        # Check for segment addresses before memory expressions
        elif state.token_type in [Tokens.COLON]:
            idx = scan_for_token(state.line, type=Tokens.REGISTER, token=X86_RE_SEGMENT_REG, ignore_type=[Tokens.SPACING], 
                                 stop_unmatched=True, match_re=True, start=-1, increment=-1)
            if idx is not None:
                state.line, state.token_type, state.token, state.orig_token = state.line[:idx], Tokens.SEGMENT_ADDRESS, \
                    state.line[idx][1] + state.token, ''.join([t[2] for t in state.line[idx:]]) + state.orig_token
                return self._handle_token(state, insert_token=False).token
        
        return state.token
    
    def handle_segment_address(self, state):
        """Handles a segment address. Defaults to returning the original token

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return state.token
    
    def handle_memory_size(self, state):
        """Handles a memory size. Removes any '{spacing}ptr' where {spacing} is any amount of spacing

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        return REPL_SPACE_PTR.sub('', state.token)
    
    def handle_memory_expression(self, state):
        """Handles memory expressions. Splits values up into 'base', 'index', 'scale', and 'displacement'

        Values:

            - 'base' (B): a register holding the starting base location. Handled by self.handle_memory_base()
            - 'index' (I): a register that is added to a base register. Handled by self.handle_memory_index()
            - 'scale' (S): an immediate value that is multiplied onto 'index', should only be 1, 2, 4, or 8. Handled by self.handle_memory_scale()
            - 'displacement' (D): and immediate value that acts as a displacement to the memory address (or in some cases,
              the literal memory address itself). Handled by self.handle_memory_displacement()
        
        Acceptable x86 memory addressing modes:

            * [D]
            * [B]
            * [B + I]
            * [B + D]
            * [B + I + D]
            * [B + I*S]
            * [I*S + D]
            * [B + I*S + D]
        
        NOTE: You can have other formats that don't fit known addressing modes, but the values might not be handled
        properly. Specifically, the first register found will be considered the 'base', and all subsequent are considered
        'index', etc.
        
        The respective handle_memory_*() methods will be called with the same parameters as this function, but 'memory_start'
        will instead be the starting index of that token, and 'token' will be the original token value (IE: before being
        handled)

        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        found_base, found_mult, found_scale = False, False, False
        for i, (token_type, new_token, old_token) in enumerate(state.line[state.memory_start:]):
            state.token_type, state.orig_token, state.token, state.token_idx = token_type, old_token, old_token, state.memory_start + i

            # Ignore spacing and whatnot
            if token_type in [Tokens.SPACING, Tokens.DISASSEMBLER_INFO, Tokens.OPEN_BRACKET, Tokens.PLUS_SIGN]:
                continue

            # If this is a register, check if it is an index or base register
            elif token_type == Tokens.REGISTER:
                if found_base:
                    new_token = self.handle_memory_index(state)
                else:
                    new_token = self.handle_memory_base(state)
                    found_base = True
            
            # If we find a times sign, then update that state
            elif token_type == Tokens.TIMES_SIGN:
                found_mult = True
                continue
            
            # If this is an immediate value, check if it should be a scale or a displacement
            elif token_type == Tokens.IMMEDIATE:
                if found_mult and not found_scale:
                    new_token = self.handle_memory_scale(state)
                    found_scale = True
                else:
                    new_token = self.handle_memory_displacement(state)
            
            state.line[state.token_idx] = (token_type, new_token, old_token)
        
        state.memory_start = None
    
    def handle_memory_base(self, state):
        """Handles the 'base' section of a memory addressing. Defaults to returning it processed as a register

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See bincfg.normalization.base_normalizer.NormalizerState
        """
        return self.handle_register(state)
    
    def handle_memory_index(self, state):
        """Handles the 'index' section of a memory addressing. Defaults to returning it processed as a register

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See bincfg.normalization.base_normalizer.NormalizerState
        """
        return self.handle_register(state)
    
    def handle_memory_scale(self, state):
        """Handles the 'scale' section of a memory addressing. Defaults to returning it processed as an immediate

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See bincfg.normalization.base_normalizer.NormalizerState
        """
        return self.handle_immediate(state)
    
    def handle_memory_displacement(self, state):
        """Handles the 'displacement' section of a memory addressing. Defaults to returning it processed as an immediate

        Should return either the token to add to the current line, or None to not add any token

        Args:
            state (NormalizerState): dictionary of current state information. See bincfg.normalization.base_normalizer.NormalizerState
        """
        return self.handle_immediate(state)

        
class X86InnerEyeNormalizer(X86BaseNormalizer):
    """A normalizer based on the Innereye method
    
    From the InnerEye paper: https://arxiv.org/pdf/1808.04706.pdf

    Rules:

        * Constant values are ignored and replaced with 'immval' or '-immval' for negative values
        * Function names are ignored and replaced with 'func'
        * Strings are 'str'
        * Jump destinations are 'immval'
        * Registers are left as-is 
        * Doesn't say anything about memory sizes, so they are ignored
        * Doesn't say anything about segment addresses, so they are ignored
        * Doesn't say anything about branch predictions, so they are ignored
        * Tokens are at the instruction-level
    
    
    Parameters
    ----------
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """"""
    renormalizable = False
    """"""

    handle_immediate = replace_immediate(include_negative=True)
    """"""
    handle_memory_size = ignore
    """"""
    handle_string_literal = replace_string_literal
    """"""
    handle_segment_address = ignore
    """"""
    handle_branch_prediction = ignore
    """"""
    opcode_function_call = replace_function_call_immediate(FUNCTION_CALL_STR)
    """"""

class X86DeepBinDiffNormalizer(X86BaseNormalizer):
    """A normalizer based on the Deep Bin Diff method
    
    From the DeepBinDiff paper: https://www.ndss-symposium.org/wp-content/uploads/2020/02/24311-paper.pdf

    Rules:

        * Constant values are ignored and replaced with 'immval'
        * General registers are renamed based on length, special ones are left as-is (with number information removed.
            EG: st5 -> st, rax -> reg8, r14d -> reg4, rip -> rip, zmm13 -> zmm)
        * Memory expressions are replaced with 'memexpr'
        * Can't really tell what's supposed to be done with function calls, will just assume they should be 'call immval'
        * Jump destinations are 'immval'
        * Strings are left as-is (Kinda bad, but they are doing binary diffing and not binary similarity, so I'll let it slide)
        * Doesn't say anything about segment addresses, so they are ignored
        * Doesn't say anything about branch predictions, so they are ignored
        * Tokens are at the op-level

    
    Parameters
    ----------
    replace_strings: `bool`
        if True, then strings will be replaced with a 'str' token. Default is False which is the default deepbindiff
        behavior in the paper and leaves full strings as individual tokens
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.OPCODE
    """"""
    renormalizable = False
    """"""

    def __init__(self, replace_strings=False, tokenizer=None, token_handlers=None, token_sep=None, tokenization_level=TokenizationLevel.AUTO, 
                 anonymize_tokens=False):
        super().__init__(tokenizer=tokenizer, token_handlers=token_handlers, token_sep=token_sep, tokenization_level=tokenization_level, anonymize_tokens=anonymize_tokens)
        
        self._replace_strings = replace_strings
        self.handle_string_literal = replace_string_literal if replace_strings else self.handle_string_literal

    
    handle_immediate = replace_immediate
    """"""
    handle_memory_size = ignore
    """"""
    handle_register = x86_replace_general_register
    """"""
    handle_memory_expression = replace_memory_expression(MEMORY_EXPRESSION_STR)
    """"""
    handle_segment_address = ignore
    """"""
    handle_branch_prediction = ignore
    """"""
    opcode_function_call = replace_function_call_immediate(IMMEDIATE_VALUE_STR)
    """"""

    def __eq__(self, other):
        return super().__eq__(other) and eq_obj(self, other, selector='._replace_strings')
    
    def __hash__(self):
        return hash_obj([super().__hash__(), self._replace_strings], return_int=True)


class X86SafeNormalizer(X86BaseNormalizer):
    """A normalizer based on the SAFE method
    
    From the SAFE paper: https://github.com/gadiluna/SAFE

    Rules:

        * All base memory addresses (IE: memory addresses that are constant values) are replaced with 'dispmem'
          NOTE: they only specify that this is used for base memory addresses. I'm not sure if they mean any displacement
          values or only those that are alone with no other registers and whatnot. It doesn't help that their implementation
          seems to have some bugs here (see below). So, I assume it is to be used for any displacement values, hence
          the string they are replaced with
        * All immediate values greater than some threshold (`safe_threshold` parameter, they use 5000 in the paper) are 
          replaced with 'immval'. Any immediate values smaller than said threshold (including those that are targets
          of call/jump instructions) are left alone
        * Memory sizes are ignored
        * Doesn't say anything about registers, so they are left as-is
        * Doesn't say anything about segment addresses, so they are left as-is
        * Doesn't say anything about branch predictions, so they are ignored
        * Strings would be ignored, just using the immediate values associated with their memory address
        * Tokens are at the instruction-level
    
    NOTE: the code from the safe paper has at least one bug I've found (specifically, in their Radare2 analyzer code which
    is the only analyzer code I could find in their repo, even though they say they also use Angr), specifically when it 
    comes to memory expressions. They do not consider some of the possible memory addressing methods that are allowed
    in x86_64 binaries. For example:

    Original Disassembly                Their Result                    Probably Intended Result
    lea esi, [esi + ecx*2]              X_lea_esi,_[esi*2+0]            X_lea_esi,_[esi+ecx*2+0]
    lea ebx, [ebx + esi*4 + 0x10]       X_lea_ebx,_[ebx*4+16]           X_lea_ebx,_[ebx+esi*4+16]
    lea edi, [eax*4 + 0x419f40]         X_lea_edi,_[MEM]                X_lea_edi,_[eax*4+MEM]

    They also seem to have problems with immediate/displacement values in those memory addresses:

    Original Disassembly                Their Result                    Probably Intended Result
    add byte [ebp + 0x4d8de455], cl     X_add_[ebp*1+1301144661],_cl    X_add_[ebp*1+MEM],_cl

    So, I take a benefit-of-the-doubt approach to this normalizer and assume the authors did not intend for this to
    happen. Instructions are normalized taking into account these possible memory addressing methods. Any displacement
    values that are larger than the safe_threshold are converted into the dispmem token.

    
    Parameters
    ----------
    imm_threshold: `int`
        immediate values whose absolute value is <= imm_threshold will be left alone, those above it will be replaced 
        with the string 'immval'
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """"""
    renormalizable = False
    """"""

    def __init__(self, imm_threshold=DEFAULT_IMMEDIATE_THRESHOLD, tokenizer=None, token_handlers=None, token_sep=None, 
                 tokenization_level=TokenizationLevel.AUTO, anonymize_tokens=False):

        self._imm_threshold = imm_threshold
        super().__init__(tokenizer=tokenizer, token_handlers=token_handlers, token_sep=token_sep, tokenization_level=tokenization_level, anonymize_tokens=anonymize_tokens)

        self.handle_immediate = threshold_immediate(self._imm_threshold)
        self.handle_memory_displacement = threshold_immediate(self._imm_threshold, DISPLACEMENT_IMMEDIATE_STR)

    handle_memory_size = ignore
    """"""
    handle_string_literal = ignore
    """"""
    handle_branch_prediction = ignore
    """"""

    def __eq__(self, other):
        return super().__eq__(other) and eq_obj(self, other, '._imm_threshold')
    
    def __hash__(self):
        return hash_obj([super().__hash__(), self._imm_threshold], return_int=True)
    

class X86DeepSemanticNormalizer(X86BaseNormalizer):
    """A normalizer based on the Deepsemantic method
    
    from the DeepSemantic paper:
    https://arxiv.org/abs/2106.05478

    Rules:

        * Immediates can fall into multiple categories:
            a. Function calls:

               - libc function name(): "libc[name]" (instead, we use "{name}")
               - recursive call: 'self'
               - function within the binary: 'innerfunc'
               - function outside the binary: 'externfunc'
               - NOTE: they do not take into account call tables that could theoretically call both inner and extern
                 functions. So, when this rather rare even occurs, it is given the token 'multifunc'

            b. Jump (branching) family: "jmpdst"

            c. Reference: (NOTE: This might not be done for all disassebly output, such as ROSE, since they don't always have
               this information readily available)
               
               - String literal: 'str'
               - Statically allocated variable: "dispbss"
               - Data (data other than a string): "dispdata"

            d. Default (all other immediate values): "immval"

        * Registers can fall into multiple categories:
            a. Stack/Base/Instruction pointer: Keep track of type and size
               [e|r]*[b|s|i]p[l]*  ->  [s|b|i]p[1|2|4|8]
            b. Special purpose (IE: flags): Keep track of type
               cr[0-15], dr[0-15], st([0-7]), [c|d|e|f|g|s]s  ->  reg[cr|dr|st], reg[c|d|e|f|s]s
            c. AVX registers: Keep track of type
               [x|y|z]*mm[0-7|0-31]  ->  reg[x|y|z]*mm
            d. General purpose registers: Keep track of size
               [e|r]*[a|b|c|d|si|di][x|l|h]*, r[8-15][b|w|d]*  ->  reg[1|2|4|8]
        
        * Pointers can fall into multiple categories:
            a. Direct, small: keep track of size
               byte,word,dword,qword,ptr  ->  memptr[1|2|4|8]
            b. Direct, large: keep track of size
               tbyte,xword,[x|y|z]mmword  ->  memptr[10|16|32|64]
            c. Indirect, string: 
               [base+index*scale+displacement]  ->  [base+index*scale+dispstr]  (NOTE: we use 'str' instead of dispstr)
            d. Indirect, not string:
               [base+index*scale+displacement]  ->  [base+index*scale+disp]

               NOTE: for our purposes, we don't necessarily always have base, index, scale, and displacement present, and
               they may appear in a different (but deterministic) order. It shouldn't really change anything for any 
               models, just how the tokens are formatted

               NOTE: it looks like the 'scale' values are their original immediate values and not replaced with the
               'immval' string, so that is taken into account as well
        
        * Tokenized at instruction-level
        * Doesn't say anything about segment addresses, so they are left as-is
        * Doesn't say anything about branch predictions, so they are ignored
    
    
    Parameters
    ----------
    special_functions: `Optional[Set[str]]`
        a set of special function names. All external functions whose name (ignoring any '@plt' at the end) is in this 
        set will have their name kept, otherwise they will be replaced with 'externfunc'. If None, will attempt to load 
        the default special function names from :func:`~bincfg.utils.cfg_utils.get_special_function_names`. If you do not
        wish to use any special function names, then pass an empty set.
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """"""
    renormalizable = False
    """"""

    def __init__(self, special_functions=None, tokenizer=None, token_handlers=None, token_sep=None, tokenization_level=TokenizationLevel.AUTO):
        self.special_functions = LIBC_FUNCTION_NAMES if special_functions is None else set(special_functions)
        super().__init__(tokenizer=tokenizer, token_handlers=token_handlers, token_sep=token_sep, tokenization_level=tokenization_level)
    
    handle_register = x86_replace_general_register
    """"""
    handle_immediate = replace_immediate
    """"""
    handle_memory_size = x86_memsize_value
    """"""
    handle_string_literal = replace_string_literal
    """"""
    handle_memory_scale = lambda self, state: str(imm_to_int(state.token))
    """"""
    handle_branch_prediction = ignore
    """"""
    opcode_function_call = special_function_call
    """"""
    opcode_jump = replace_jump_destination
    """"""

    def __eq__(self, other):
        return super().__eq__(other) and eq_obj(self, other, selector='special_functions')
    
    def __hash__(self):
        return hash_obj([super().__hash__(), self.special_functions], return_int=True)


class X86CompressedStatsNormalizer(X86BaseNormalizer):
    """A normalizer I created for use in CFG.get_compressed_stats()

    Rules:

        * Immediates are replaced with immediate string (including negative)
        * function calls are either self vs. intern vs. extern func, no special functions
        * jump destinations are 'jmpdst'
        * registers are handled the same as deepsem/deepbindiff
        * memory pointers/memory expressions are handled the same as in deepsemantic
        * Tokenized at the instruction-level
        * segment addresses are ignored
        * branch predictions are ignored
    
    Parameters
    ----------
    special_functions: `Optional[Set[str]]`
        a set of special function names. All external functions whose name (ignoring any '@plt' at the end) is in this 
        set will have their name kept, otherwise they will be replaced with 'externfunc'. If None, will attempt to load 
        the default special function names from :func:`~bincfg.utils.cfg_utils.get_special_function_names`. If you do not
        wish to use any special function names, then pass an empty set.
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """"""
    renormalizable = False
    """"""

    def __init__(self, special_functions=None, tokenizer=None, token_handlers=None, token_sep=None, tokenization_level=TokenizationLevel.AUTO):

        self.special_functions = set() if special_functions is None else special_functions
        super().__init__(tokenizer=tokenizer, token_handlers=token_handlers, token_sep=token_sep, tokenization_level=tokenization_level)
            
    handle_register = x86_replace_general_register
    """"""
    handle_memory_size = x86_memsize_value
    """"""
    handle_immediate = replace_immediate(include_negative=True)
    """"""
    handle_string_literal = replace_string_literal
    """"""
    handle_segment_address = ignore
    """"""
    handle_branch_prediction = ignore
    """"""
    opcode_function_call = special_function_call
    """"""
    opcode_jump = replace_jump_destination
    """"""

    def __eq__(self, other):
        return super().__eq__(other) and eq_obj(self, other, selector='special_functions')
    
    def __hash__(self):
        return hash_obj([super().__hash__(), self.special_functions], return_int=True)


class X86HPCDataNormalizer(X86BaseNormalizer):
    """A special normalizer meant for use in HPC compile jobs
    
    This normalizer is made to reduce the total number of new tokens as much as possible while still being able to fully
    reproduce original BaseNormalizer output, and while trying to minimize the number of tokens per assembly line as
    much as possible.

    Since immediate values make up the vast majority of 'unique' tokens (or are the root cause of there being so many)
    in the BaseNormalizer, this is all that is changed. Specifically:

        - immediate values get split into multiple tokens. EG: '123456789' -> '1234', '5678', '9' if using `num_digits` of 4
        - negatives stay connected to tokens. EG: '-54321' -> '-543', '21' if using `num_digits` of 4
        - Before some split immediate values, a 'split immediate' token is inserted for later tokenization to know that the
          following immediate values should all be concatenated together. This is only inserted when default behavior
          would produce the wrong values (EG: whenever a token is split, whenever a non-split token has a split token
          before it, etc.)
    
    NOTE: this should only be used with the 'opcode' tokenization level as it provides no benefit otherwise
    

    Parameters
    ----------
    num_digits: `int`
        the number of digits to use before splitting. This will include the minus sign as a digit.
    replace_strings: `bool`
        if True, then strings will be replaced with a 'str' token. Default is False which is the default deepbindiff
        behavior in the paper and leaves full strings as individual tokens
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.OPCODE
    """"""

    @property
    def renormalizable(self):
        """We are only losslessly renormalizable if self._replace_strings == False"""
        return hasattr(self, '_replace_strings') and not self._replace_strings

    def __init__(self, num_digits=4, replace_strings=False, tokenizer=None, token_handlers=None, token_sep=None, tokenization_level=TokenizationLevel.AUTO):
        if num_digits < 1:
            raise ValueError("`num_digits` must be >= 1")
        
        self._num_digits = num_digits
        self._replace_strings = replace_strings
        super().__init__(tokenizer=tokenizer, token_handlers=token_handlers, token_sep=token_sep, tokenization_level=tokenization_level)

        self.handle_string_literal = replace_string_literal(replace_previous_immediate=True) if replace_strings else super().handle_string_literal

    def finalize_instruction(self, state):
        """Handles a single instruction. Calls super()'s handle_instruction, then performs the immediate splitting
        
        Args:
            state (NormalizerState): dictionary of current state information. See ``bincfg.normalization.base_normalizer.NormalizerState``
        """
        super().finalize_instruction(state)

        new_inst = []
        prev_split = False
        for token_type, new_token, old_token in state.line:
            if token_type in [Tokens.IMMEDIATE] and (len(new_token) > self._num_digits or prev_split):
                
                new_inst.append((Tokens.SPLIT_IMMEDIATE, SPLIT_IMMEDIATE_TOKEN, SPLIT_IMMEDIATE_TOKEN))
                while len(new_token) > 0:
                    new_inst.append((Tokens.IMMEDIATE, new_token[:self._num_digits], new_token[:self._num_digits]))
                    new_token = new_token[self._num_digits:]

                prev_split = True
            else:
                new_inst.append((token_type, new_token, old_token))
                prev_split = False
        
        state.line = new_inst

    def __eq__(self, other):
        return super().__eq__(other) and eq_obj(self, other, selector='._num_digits') and eq_obj(self, other, selector='._replace_strings')
    
    def __hash__(self):
        return hash_obj([super().__hash__(), self._num_digits, self._replace_strings], return_int=True)
    


import warnings
import bincfg
from collections import Counter
from .cfg_basic_block import CFGBasicBlock
from ..utils import eq_obj, hash_obj, get_address
from ..utils.type_utils import *


if IN_PYTHON_TYPING_VERSION:
    CFGFunctionPickledState = Tuple[int, str, Tuple[CFGBasicBlock, ...], bool]
    """The pickled state of a function"""


class CFGFunction:
    """A single function in a ``CFG``

    Can be initialized empty, or by passing kwarg values.

    NOTE: these objects should not be pickled/copy.deepcopy()-ed by themselves, only as a part of a cfg

    Parameters
    ----------
    parent_cfg: `Optional[bincfg.CFG]`
        the parent ``CFG`` object to which this ``CFGFunction`` belongs
    address: `Optional[AddressLike]`
        the memory address of this function. If not present, then the address will be set to -1
    name: `Optional[str]`
        the string name of this function. If not present, or if the name passed is the empty string, this function is 
        given a default name '__UNNAMMED_FUNC_X' where 'X' is the memory address
    blocks: `Optional[Iterable[CFGBasicBlock]]`
        if None, will be initialized to an empty list, otherwise an iterable of ``CFGBasicBlock`` objects that are within 
        this function
    is_extern_function: `bool`
        if True, then this function is an external function (a dynamically loaded function)
    metadata: `Optional[dict]`
        optional dictionary of metadata to associate with this function
    """

    parent_cfg: 'Union[bincfg.CFG, None]'
    """the parent ``CFG`` object to which this ``CFGFunction`` belongs, or None if it hasn't been initialized yet"""

    address: 'int'
    """the integer memory address of this function. Will be -1 if not initialized yet"""

    name: 'str'
    """the string name of this function. Will be given a default name based on its memory address if not present"""

    blocks: 'list[CFGBasicBlock]'
    """list of all basic blocks in this function"""

    metadata: 'dict'
    """Dictionary of metadata associated with this function"""

    def __init__(self, parent_cfg: 'Optional[bincfg.CFG]' = None, address: 'Optional[AddressLike]' = None, name: 'Optional[str]' = None, 
                 blocks: 'Optional[Iterable[CFGBasicBlock]]' = None, is_extern_function: bool = False, metadata: 'Optional[dict]' = None):
        self.parent_cfg = parent_cfg
        self.address = -1 if address is None else get_address(address)
        self.name = name if name is not None and name != '' else ("__UNNAMED_FUNC_%d" % self.address)
        self.blocks = [] if blocks is None else list(blocks)
        self.metadata = {} if metadata is None else metadata
        self._is_extern_function: bool = is_extern_function
    
    @property
    def num_blocks(self) -> 'int':
        """The number of basic blocks in this function"""
        return len(self.blocks)
    
    @property
    def num_asm_lines(self) -> 'int':
        """The total number of assembly lines across all blocks in this function"""
        return sum(b.num_asm_lines for b in self.blocks)
    
    @property
    def asm_counts(self) -> 'Mapping[str, int]':
        """A ``collections.Counter`` of all unique assembly lines and their counts in this function"""
        return sum((b.asm_counts for b in self.blocks), Counter())
    
    @property
    def is_root_function(self) -> 'bool':
        """True if this function is not called by any other functions, False otherwise"""
        return len(self.called_by) == 0
    
    @property
    def is_recursive(self) -> 'bool':
        """True if this function calls itself at some point
        
        Specifically, if at least one ``CFGBasicBlock`` in this ``CFGFunction.blocks`` list has an `edges_out` function
        call address that is equal to this ``CFGFunction``'s address
        """
        return self.address is not None and any(block.calls(self.address) for block in self.blocks)
    
    @property
    def is_extern_function(self) -> 'bool':
        """True if this function is an external function, False otherwise"""
        return self._is_extern_function
    
    @property
    def is_intern_function(self) -> 'bool':
        """True if this function is an internal function, False otherwise"""
        return not self._is_extern_function
    
    @property
    def function_entry_block(self) -> 'CFGBasicBlock':
        """The ``CFGBasicBlock`` that is the function entry block
        
        Specifically, returns the first ``CFGBasicBlock`` found that has the same address as this function (there ~should~
        only be one as each basic block ~should~ have a unique memory address)
        """
        for block in self.blocks:
            if block.address == self.address:
                return block
        raise ValueError("Function %s does not have an entry block!" % (repr(self.name) if self.name is not None else 'with no name'))
    
    @property
    def called_by(self) -> 'list[CFGBasicBlock]':
        """A list of ``CFGBasicBlock``'s that call this function
        
        Specifically, the list of all ``CFGBasicBlock`` objects in this function's `.parent_cfg` CFG object that call 
        this function. If this ``CFGFunction`` has no parent, then the empty list will be returned.
        
        NOTE: this is computed dynamically each call (as ``CFG`` objects are mutable), so it may be useful to compute it
        once per function and save it if needed
        """
        if self.parent_cfg is None:
            return list()

        return [block for block in self.parent_cfg.blocks if block.calls(self)]
    
    def __str__(self) -> 'str':
        un_funcs = set([self.parent_cfg.get_block(b.address).parent_function.address for b in self.called_by])
        include_self = " (including self)" if self.is_recursive else ""
        extra_str = ", called by %d basic blocks across %d functions%s" % (len(self.called_by), len(un_funcs), include_self)
        addr_str = ("0x%08x" % self.address) if self.address is not None else "NO_ADDRESS"

        func_name_str = ("'%s'" % self.name) if self.name is not None else "with no name"
        num_asm_lines = sum([len(b.asm_lines) for b in self.blocks])
        return "CFGFunction %s %s at %s with %d blocks, %d assembly lines%s with metadata: %s" \
            % ('externfunc' if self.is_extern_function else 'innerfunc', func_name_str, addr_str, len(self.blocks), 
                num_asm_lines, extra_str, self.metadata)
    
    def __repr__(self) -> 'str':
        return str(self)
    
    def __eq__(self, other) -> 'bool':
        if not isinstance(other, CFGFunction) or len(self.blocks) != len(other.blocks) or \
            not all(eq_obj(self, other, selector=s) for s in ['address', 'name', 'is_extern_function', 'metadata']):
            return False
        
        # Check all the blocks are the same using a dictionary since sets don't work for some reason...
        d_self, d_other = {}, {}
        for blocks, count_dict in [(self.blocks, d_self), (other.blocks, d_other)]:
            for b in blocks:
                _hash = hash(b)
                if _hash not in count_dict:
                    count_dict[_hash] = (1, b)
                else:
                    assert b == count_dict[_hash][1]
                    count_dict[_hash] = (count_dict[_hash][0] + 1, b)
        
        return eq_obj(d_self, d_other)
    
    def __hash__(self) -> 'int':
        return hash_obj([self.name, self.address, self.blocks, self._is_extern_function, self.metadata], return_int=True)

    def __getstate__(self) -> 'CFGFunctionPickledState':
        """Print a warning about pickling singleton function objects"""
        warnings.warn("Attempting to pickle a singleton function object! This will mess up edges unless you know what you're doing!")
        return self._get_pickle_state()
    
    def __setstate__(self, state: 'CFGFunctionPickledState') -> None:
        """Print a warning about pickling singleton function objects"""
        warnings.warn("Attempting to unpickle a singleton function object! This will mess up edges unless you know what you're doing!")
        self._set_pickle_state(state)
    
    def _get_pickle_state(self) -> 'CFGFunctionPickledState':
        """Returns info of this CFGFunction as a tuple"""
        return (self.address, self.name, tuple(b._get_pickle_state() for b in self.blocks), self.is_extern_function)

    def _set_pickle_state(self, state: 'CFGFunctionPickledState'):
        """Sets state from _get_pickle_state"""
        self.address, self.name, blocks, self._is_extern_function = state
        self.blocks = [CFGBasicBlock(parent_function=self)._set_pickle_state(b) for b in blocks]
        return self



"""A Python tool designed to parse binary analyzer outputs to produce call flow graphs (CFG), tokenize and normalize the assembly lines within those CFGs, and convert that data into ML-ready formats."""

__version__ = '1.0.0'
__author__ = 'Justin Allen'

from .utils import *  # Keep this first!
from .cfg import *
from .normalization import *



import numpy as np
import pickle
import traceback
from .cfg_dataset import CFG, CFGDataset
from .mem_cfg import MemCFG
from ..utils import check_for_normalizer, isinstance_with_iterables, eq_obj, hash_obj, AtomicTokenDict
from ..normalization import get_normalizer, normalize_cfg_data


class MemCFGDataset:
    """A CFGDataset that is more memory efficient
    
    Parameters
    ----------
    cfg_data: `Optional[Union[str, CFG, CFGDataset, MemCFG, MemCFGDataset, Iterable]]`
        the data to use. Can be None for an empty dataset, or a string (for input to CFG), `CFG`, `CFGDataset`, `MemCFG`, 
        `MemCFGDataset`, or iterable of those values to add that data to this dataset
    tokens: `Optional[Union[Dict[str, int], AtomicTokenDict]]`
        if passed, will initialize the token dictionary to this dictionary of tokens (will be copied). Can be an
        AtomicTokenDict to use an atomic file token dictionary
    normalizer: `Optional[Union[str, Normalizer]]`
        the normalizer to use, or None to default to the normalizer of the first added CFG/MemCFG
    metadata: `Optional[Dict]`
        a dictionary of metadata to attach to this MemCFGDataset
        NOTE: passed dictionary will be shallow copied
    add_data_kwargs: `Any`
        kwargs to pass to self.add_data() when adding the passed cfg_data
    """

    cfgs = None
    """The list of all memcfgs in this dataset"""

    normalizer = None
    """The normalizer used in this dataset, or None if there is no normalizer"""

    metadata = None
    """A dictionary of metadata associated with this ``MemCFGDataset``"""

    using_tokens = None
    """A dictionary mapping string tokens to their integer values
    
    Can be an AtomicTokenDict for atomic updates to tokens
    """

    def __init__(self, cfg_data=None, using_tokens=None, normalizer=None, metadata=None, **add_data_kwargs):
        self.cfgs = []
        self.tokens = {} if using_tokens is None else using_tokens
        self.normalizer = get_normalizer(normalizer) if normalizer is not None else None
        self.metadata = {} if metadata is None else metadata.copy()
        self._curr_cfg_memory_usage = 0

        if cfg_data is not None:
            self.add_data(cfg_data, **add_data_kwargs)
        elif len(add_data_kwargs) != 0:
            raise ValueError("Passed add_data_kwargs, but no data to add: %s" % add_data_kwargs)
    
    def add_data(self, *cfg_data, inplace=True, force_renormalize=False, progress=False):
        """Adds data to this dataset

        Args:
            cfg_data (Union[str, CFG, MemCFG, CFGDataset, MemCFGDataset, Iterable]): arbitrary amount of 
                str (CFG input)/CFG/MemCFG/CFGDataset/MemCFGDataset's, or iterables of them, to add to this dataset
            inplace (bool, optional): whether or not to normalize the incoming cfg_data inplace. Defaults to True.
            force_renormalize (bool, optional): by default, this method will only normalize cfg's whose 
                .normalizer != to this dataset's normalizer. However if `force_renormalize=True`, then all cfg's will be
                renormalized even if they have been previously normalized with the same normalizer. Defaults to False.
            mp (bool, optional): if True, will use multiprocessing to normalize cfgs. Defaults to False.
            progress (bool, optional): if True, will show a progressbar when adding multiple cfgs. Defaults to False.

        Raises:
            TypeError: if something other than a cfg/dataset is passed in `cfg_data`
        """
        # Check that all elements in cfg_data are CFG's/CFGDataset's/MemCFG's/MemCFGDataset's, or iterables of them
        temp = []
        for cfg in cfg_data:
            try:
                temp += isinstance_with_iterables(cfg, (str, CFG, CFGDataset, MemCFG, MemCFGDataset), recursive=False, ret_list=True)
            except:
                raise TypeError("Can only add str's/CFG's/CFGDataset's/MemCFG's/MemCFGDataset's, or iterables of them, to "
                    "CFGDataset, not '%s'" % type(cfg).__name__)
        cfg_data = temp

        # Check to make sure we have a normalizer to use
        if self.normalizer is None:
            check_for_normalizer(self, cfg_data)
        
        cfg_data = normalize_cfg_data(cfg_data, normalizer=self.normalizer, inplace=inplace, using_tokens=self.tokens,
            force_renormalize=force_renormalize, convert_to_mem=True, unpack_cfgs=True, progress=progress)
        
        # Set the cfg's normalizer and tokens attributes to this dataset's objects
        for cfg in cfg_data:
            cfg.normalizer, cfg.tokens = self.normalizer, self.tokens

        self.cfgs += cfg_data
    
    def normalize(self, normalizer=None, inplace=True, force_renormalize=False, progress=False):
        """Normalize this ``MemCFGDataset``.

        Args:
            normalizer (Union[str, Normalizer]): the normalizer to use. Can be a ``Normalizer`` object, or a 
                string, or None to use the default BaseNormalizer(). Defaults to None.
            inplace (bool, optional): by default, normalizes this dataset inplace (IE: without copying objects). Can set
                to False to return a copy. Defaults to True.
            force_renormalize (bool, optional): by default, this method will only normalize cfg's whose 
                .normalizer != to the passed normalizer. However if `force_renormalize=True`, then all cfg's will be 
                renormalized even if they have been previously normalized with the same normalizer.. Defaults to False.
            progress (bool, optional): if True, will show a progressbar while normalizing. Defaults to False.

        Returns:
            MemCFGDataset: this dataset normalized
        """
        # We don't want to pass using_tokens because we are normalizing with new tokens (most likely)
        return normalize_cfg_data(self, normalizer, inplace=inplace, force_renormalize=force_renormalize, progress=progress)
        
    def remove_cfg(self, cfg_or_idx):
        """Removes the given MemCFG (or index of MemCFG if cfg_or_idx is an integer) from this MemCFGDataset

        Args:
            cfg_or_idx (Union[MemCFG, int]): cfg or index to remove
        """
        if isinstance(cfg_or_idx, (int, np.integer)):
            cfg_or_idx = self.cfgs[cfg_or_idx]
        
        self.cfgs.remove(cfg_or_idx)
    
    @property
    def num_cfgs(self):
        return len(self.cfgs)

    @property
    def num_blocks(self):
        return sum(cfg.num_blocks for cfg in self.cfgs)
    
    @property
    def num_asm_lines(self):
        return sum(cfg.num_asm_lines for cfg in self.cfgs)
    
    @property
    def num_edges(self):
        return sum(cfg.num_edges for cfg in self.cfgs)
    
    @property
    def num_functions(self):
        return sum(cfg.num_functions for cfg in self.cfgs)

    def save(self, path, freeze_tokens=True):
        """Saves this MemCFGDataset to path
        
        Args:
            path (str): the filepath to save to
            freeze_tokens (bool): whether or not to 'freeze' the tokens in this MemCFGDataset. 'freezing' the tokens
                just means that, if an AtomicTokenDict is the current token dictionary for this MemCFGDataset, then
                its current data will be saved in the pickle file as a normal dict. This is useful for loading this
                data later so that the loading does not depend on being able to access the files for the AtomicTokenDict.
                Default: True. If the token dictionary is already a dict, then this has no effect
        """
        old_tokens = self.tokens
        self.tokens = self.tokens.data if isinstance(self.tokens, AtomicTokenDict) and freeze_tokens else self.tokens
        with open(path, 'wb') as f:
            pickle.dump(self, f)
        self.tokens = old_tokens
    
    def dumps(self):
        """Returns this object pickled with pickle.dumps()"""
        return pickle.dumps(self)
    
    @classmethod
    def load(cls, path):
        """Loads this MemCFGDataset from path"""
        try:
            with open(path, 'rb') as f:
                return pickle.load(f)
        except Exception as e:
            raise ValueError("Error reading pickle file: %s. Reason:\n%s: %s\nTraceback: %s" 
                             % (repr(path), type(e).__name__, e, traceback.format_exc()))
    
    def __getstate__(self):
        """State for pickling"""
        state = self.__dict__.copy()
        state['cfgs'] = [{k: v for k, v in cfg.__getstate__().items() if k not in ['normalizer', 'tokens']} for cfg in state['cfgs']]
        return state
    
    def __setstate__(self, state):
        """State for unpickling"""        
        # Setting attributes, and cfgs list
        self.cfgs = []
        for k, v in state.items():
            if k == 'cfgs':
                for cfg_dict in v:
                    new_cfg = MemCFG(None)
                    new_cfg.__setstate__(cfg_dict)
                    self.cfgs.append(new_cfg)
            else:
                setattr(self, k, v)
        
        # Updating the normalizer/tokens in the cfgs
        for cfg in self.cfgs:
            cfg.normalizer = self.normalizer
            cfg.tokens = self.tokens
    
    def __eq__(self, other):
        return isinstance(other, MemCFGDataset) and all(eq_obj(self, other, selector=s) for s in 
            ['normalizer', 'tokens', 'metadata', 'cfgs'])
    
    def __hash__(self):
        return sum(hash(c) for c in self.cfgs) * 11 + hash_obj([self.tokens, self.metadata], return_int=True) * 13

    def __str__(self):
        return "MemCFGDataset with %d cfg's and %s tokens" % (len(self.cfgs), len(self.tokens) if self.tokens is not None else 'UNKNOWN')
    
    def __str__(self):
        stat_names = ["CFG's", 'Functions', 'Edges', 'Basic Blocks', 'Assembly Lines']
        c = [self.num_cfgs, self.num_functions, self.num_edges, self.num_blocks, self.num_asm_lines]
        stats = _get_stats(stat_names, c)
        norm = ('with normalizer: %s' % self.normalizer) if self.normalizer is not None else 'with no normalizer'
        return "%s %s\nStats:\n%s" % (self.__class__.__name__, norm, stats)
    
    def __repr__(self):
        return self.__str__()
    
    def __len__(self):
        return len(self.cfgs)
    
    def __getitem__(self, idx):
        return self.cfgs[idx]
    
    def __iter__(self):
        return iter(self.cfgs)


def _get_stats(stat_names, counts):
    """Returns a nicely-printable set of statistics for the MemCFGDataset

    Args:
        stat_names (Iterable[str]): the names for each statistic
        counts (Iterable[int]): the values for each statistic
    
    Returns:
        str: the nicely formatted set of statistics for the MemCFGDataset
    """
    return '\n'.join([('\t' + name + ': ' + str(c)) for name, c in zip(stat_names, counts)])



from ..base_tokenizer import Tokens, BaseTokenizer, Architectures
from ..norm_utils import RE_SPACING, RE_NEWLINE, scan_for_token
import re


# The delimiter lookahead. Allows for lookahead to make sure there is some delimiter, and we don't greedily grab 
#   registers/values/etc in case their names interfere with anything else.
# Can be a spacing, newline, or end of string
X86_RE_DELIMITER = r'{spacing}|{newline}|[\[\]+*<>:]|$'.format(spacing=RE_SPACING, newline=RE_NEWLINE)

# Dictionary mapping all x86_64 registers to their sizes in bytes
X86_REGISTER_SIZES = {
    k:v for a, v in [
        (['xmm%d' % i for i in range(16)], 16),                             # XMM 16-byte registers
        (['ymm%d' % i for i in range(16)], 32),                             # YMM 32-byte registers
        (['zmm%d' % i for i in range(32)], 64),                             # ZMM 64-byte registers

        (['st%d' % i for i in range(16)], 10),                              # ST 10-byte registers (rose interpretation)
        (['st(%d)' % i for i in range(16)], 10),                            # ST 10-byte registers (ghidra interpretation)
        (['st'], 10),                                                       # Plain 'st' is just 'st(0)'
        (['mm%d' % i for i in range(8)], 8),                                # MM 8-byte registers [lower 64 bits of ST]
        
        (['cw', 'fp_ip', 'fp_dp', 'fp_cs', 'sw', 'tw', 
            'fp_ds', 'fp_opc', 'cs', 'ss', 'ds', 'es', 'fs', 
            'gs', 'gdtr', 'idtr', 'ldtr', 'tr', 'msw'], 2),                 # Various special 2-byte registers
        (['fp_dp', 'fp_ip', 'mxcsr'], 4),                                   # Various special 4-byte registers

        (['%sr%d' % (k, v) for k in ['c', 'd'] for v in range(16)], 8),     # DR(0-15)/CR(0-15) 8-byte registers

        (['r%d' % v for v in range(8, 16)], 8),                             # R(8-15)  8-byte registers
        (['r%dd' % v for v in range(8, 16)], 4),                            # R(8-15)d 4-byte registers
        (['r%dw' % v for v in range(8, 16)], 2),                            # R(8-15)w 2-byte registers
        (['r%db' % v for v in range(8, 16)], 1),                            # R(8-15)b 1-byte registers

        (['rax', 'rbx', 'rcx', 'rdx'], 8),                                  # Named general 8-byte r-registers
        (['eax', 'ebx', 'ecx', 'edx'], 4),                                  # Named general 4-byte r-registers
        (['ax', 'bx', 'cx', 'dx'], 2),                                      # Named general 2-byte r-registers

        (['rflags', 'rip', 'rbp', 'rsi', 'rdi', 'rsp'], 8),                 # Various special 8-byte r-registers
        (['eflags', 'eip', 'ebp', 'esi', 'edi', 'esp'], 4),                 # Various special 4-byte e-registers
        (['flags', 'ip', 'bp', 'si', 'di', 'sp'], 2),                       # Various special 2-byte registers

        (['bpl', 'sil', 'dil', 'spl'], 1),                                  # Various special 1-byte registers
        (['%s%s' % (a, b) for a in 'abcd' for b in 'lh'], 1),               # Various general 1-byte registers (low and high)
    ]
    for k in a
}

# Dictionary mapping all base memory sizes to their size in bytes
# NOTE: Should find someone who can confirm sizes here, because I get different values when googling what 'ldouble' is...
X86_MEMORY_SIZES = {
    'byte': 1, 'word': 2, 'dword': 4, 'qword': 8, 'tword': 10, 'float': 4, 'tfloat': 10, 'double': 8, 'ldouble': 16,
    'xmmword': 16, 'ymmword': 32, 'zmmword': 64,
}

# Regex matches to registers. This is ~15% faster than brute-force matching all keys in REGISTER_SIZES
# Registers can be followed by a '*', ']', or delimiter
X86_RE_REG_RE_MATCH = r'[xyz]?mm[0-9]+|st\(?[0-9]*\)?|(?:[sb]p|[ds]i)l|[re]?(?:flags|ip|[bs]p|[sd]i|[abcd]x)|[cd]r[0-9]+|r[0-9]+[dwb]?|[abcd][lh]|[cst]w|fp_(?:[id]p|[cd]s|opc)|[cdefgs]s|(?:[gil]d)?tr|msw|mxcsr'
X86_RE_ALL_REGISTERS = r'(?:{all_reg})(?={delim})'.format(all_reg='|'.join([X86_RE_REG_RE_MATCH]), delim=X86_RE_DELIMITER)

# x86_64 instruction prefixes, stolen from: http://web.mit.edu/rhel-doc/3/rhel-as-en-3/i386-prefixes.html.
X86_RE_INSTRUCTION_PREFIX = r'(?:lock|rep(?:ne|nz|e|z)?)(?={delim})'.format(delim=X86_RE_DELIMITER)

# For branch prediction
X86_RE_BRANCH_PREDICTION = r'p[tn](?={delim})'.format(delim=X86_RE_DELIMITER)

# Known memory sizes. Includes a 'v%d' in front of all of them to denote a vector of multiple of these. Optional 'ptr' after spacing
X86_RE_MEM_SIZES = r'(?:v[0-9]+)?(?:byte|[dqt]?word|t?float|l?double|[xyz]mmword)(?:{spacing}ptr)?(?={delim})'.format(delim=X86_RE_DELIMITER, spacing=RE_SPACING)

# Check for any opcode mnemonic. For now, we just get any and all characters that could correspond to an opcode, 
#   instruction prefix, or branch prediction. These will be sorted out later when parsing. Must be followed by delimiter
X86_RE_OPCODE = r'(?:[a-z][a-z0-9_]*)(?={delim})'.format(delim=X86_RE_DELIMITER)

# For finding segment registers, which are handled only in the normalizer
X86_RE_SEGMENT_REG = r'[cdefgs]s'

# Match names to tokens, and define the order in which they should be matched
X86_DEFAULT_TOKENS = [
    (Tokens.INSTRUCTION_PREFIX, X86_RE_INSTRUCTION_PREFIX),  # Instructions prefixes (EG: lock, rep, etc.)
    (Tokens.MEMORY_SIZE, X86_RE_MEM_SIZES),  # Known memory sizes. Must go before reg/opcode
    (Tokens.REGISTER, X86_RE_ALL_REGISTERS),  # All known registers accross all architectures
    (Tokens.BRANCH_PREDICTION, X86_RE_BRANCH_PREDICTION),  # The 'pt' and 'pn' branch predictions
    (Tokens.OPCODE, X86_RE_OPCODE),  # Opcode. Should go last since it's the most general matching. 
]
"""Default list of (token_type, regex) token tuples to match to"""


class X86BaseTokenizer(BaseTokenizer):
    """A default class to tokenize x86 assembly line input

    This class matches the following tokens:

        * All of the default special tokens from parent (see :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer`)
        * Instruction prefix tokens (EG: 'lock', 'repe', etc.)
        * Memory sizes (EG: 'qword', 'byte ptr', 'xmmword', 'v2float', etc.)
        * Registers (see ``bincfg.normalization.x86.x86_tokenizer.X86_REGISTER_SIZES`` for the list of them)
        * Prepended or appended instruction prefixes and branch predictions. You can prepend or append either of those
          to opcodes while separating with either ',', '_', or '.'. EG: "lock_str", "jnz,pt". This will only apply to
          known prefixes and branch prediction strings; if it is unknown, it is considered a part of a larger opcode
          (EG: "vcmpneq_oqss").

          Known prefixes: ['lock', 'rep', 'repe', 'repz', 'repne', 'repnz']
          Known branch predictions: ['pt', 'pn']

    This will perform the following transformations to the incomming token stream:

        * Instruction prefixes and branch prediction tokens may be reordered to keep ordering consistent. NOTE: if you
          do not wish to have this behavior, pass 'reorder_tokens=False' as a kwarg to the `.tokenize()` call.

          Any instruction prefixes alone on their own line will be moved to the next line under the assumption that is
          the opcode they are affecting (no check is done to ensure the first token in the subsequent line is actually
          an opcode, however). Then, for each opcode, we get all surrounding instruction prefix and branch prediction
          tokens (ignoring spacing, grabbing all tokens until reaching a non-instruction prefix and non-branch prediction
          token). These are reordered such that all instruction prefixes come first, then branch predictions after those,
          and finally the opcode token. They will appear in the order that they were found in the token list. There will
          only be a single space ' ' acting as spacing inbetween them.


    Parameters
    ----------
    tokens: `Optional[List[Tuple[str, str]]]`
        the tokens to use. Should be a list of 2-tuples. Each tuple is a pair of (name, regex) where
        name is the string name of the token, and regex is a regular expression to find that token. These
        tuples should be ordered in the preferred order to search for tokens. If None, then this will default to 
        self.DEFAULT_TOKENS, which is ``bincfg.normalization.x86.x86_tokenizer.X86_DEFAULT_TOKENS``
    token_handlers: `Optional[Dict[str, Callable[[Dict[str, Any]], Union[None, str]]]]`
        optional dictionary mapping token type strings to functions to handle those token types when tokenizing. This is
        intended to be used when you wish to add entirely new token types not present in `bincfg.normalization.base_tokenizer.Tokens`.
        If you wish to change the behavior of handling an already-present token type, just override that token handler function.
        These will override the default token handlers.
    insert_special_tokens: `bool`
        by default, some special tokens will be inserted at the front of `tokens` (see the 'special tokens' listed above).
        If you wish to stop this from happening, you can set `insert_special_tokens` to False
    case_sensitive: `bool`
        If True, then regular expressions will be matched exactly as they appear. If False, then the re.IGNORECASE flag
        will be passed when compiling the regular expressions
    """
    DEFAULT_TOKENS = X86_DEFAULT_TOKENS

    ARCHITECTURE = Architectures.X86
    """The architecture this tokenizer works on"""


X86_DEFAULT_TOKENIZER = X86BaseTokenizer()



from .base_tokenizer import Tokens
from .base_normalizer import DEFAULT_IMMEDIATE_THRESHOLD
from .norm_utils import imm_to_int, scan_for_token, DISPLACEMENT_IMMEDIATE_STR, IMMEDIATE_VALUE_STR, STRING_LITERAL_STR, \
    JUMP_DESTINATION_STR, MEMORY_EXPRESSION_STR, FUNCTION_CALL_STR, MULTI_FUNCTION_CALL_STR, RECURSIVE_FUNCTION_CALL_STR, \
    EXTERNAL_FUNCTION_CALL_STR, INTERNAL_FUNCTION_CALL_STR
from ..cfg.cfg_edge import EdgeType
import bincfg


def ignore(self, state):
    """Ignores information (if using for rose info, then it will also ignore negatives)"""
    pass


def identity(self, state):
    """Returns the original token"""
    return state.token


def return_token(self, state):
    """Returns the original token"""
    return state.token


def return_dispmem(self, state):
    """Replaces memory addressing displacement values with the string 'dispmem'"""
    return DISPLACEMENT_IMMEDIATE_STR


def replace_immediate(*args, include_negative=False):
    """Builds a function that replaces immediate values with the IMMEDIATE_VALUE_STR. 
    
    This will return a function to be called as a part of a normalizer. This function takes no arguments and only 1 keyword 
    argument: whether or not to include a negative sign '-' in front of the immediate string when the input is negative.

    NOTE: This is meant to be a higher-order function. But, just in case the user forgets that (or is too lazy to add in
    two extra characters to call this function), if you pass multiple args then it will be assumed this is being called 
    as if it is the _repl_func() function below and will simply return the default result

    Args:
        args: args for this function. Ideally empty
        include_negative (bool, optional): if True, will include a negative sign in front of the returned immediate 
            string when the input is negative. Defaults to False.

    Returns:
        Union[Callable[..., str], str]: either a function that will handle immediate strings (if this function was 
            called correctly), or a handled immediate string
    """
    def _ret_imm(self, state):
        return ('-' + IMMEDIATE_VALUE_STR) if include_negative and imm_to_int(state.token) < 0 else IMMEDIATE_VALUE_STR
    return _ret_imm if len(args) == 0 else _ret_imm(*args)


def replace_string_literal(*args, replace_previous_immediate=False):
    """Builds a function that replaces string literal values with the string 'str'
    
    This will return a function to be called as a part of a normalizer. This function takes no arguments and only 1 keyword 
    argument: whether to replace the previous immediate, or keep it and add in a 'str' string

    NOTE: This is meant to be a higher-order function. But, just in case the user forgets that (or is too lazy to add in
    two extra characters to call this function), if you pass multiple args then it will be assumed this is being called 
    as if it is the _repl_func() function below and will simply return the default result

    Args:
        args: args for this function. Ideally empty
        replace_previous_immediate (bool): if True, then any previous immediate value will be replaced with the 'str'
            string, otherwise the 'str' string will just be added

    Returns:
        Union[Callable[..., str], str]: either a function that will handle immediate strings (if this function was 
            called correctly), or a handled immediate string
    """
    def _ret_string_imm(self, state):
        # If we are not replacing any previous immediate, just return now
        if not replace_previous_immediate:
            return STRING_LITERAL_STR
        
        # Scan for a previous immediate token, and if not present, just return the string now
        idx = scan_for_token(state.line, type=Tokens.IMMEDIATE, ignore_type=[Tokens.SPACING], stop_unmatched=True, start=-1, increment=-1)
        if idx is None:
            return STRING_LITERAL_STR
        
        # We have a previous immediate value, replace it with our string, and return None to not insert token
        state.line = state.line[:idx] + [(Tokens.STRING_LITERAL, STRING_LITERAL_STR, state.orig_token)]
        return None
    return _ret_string_imm if len(args) == 0 else _ret_string_imm(*args)


def threshold_immediate(threshold=DEFAULT_IMMEDIATE_THRESHOLD, include_negative=False, imm_str=IMMEDIATE_VALUE_STR):
    """Builds a function that replaces immediate values with `immval` iff abs(immediate) > some threshold

    Args:
        threshold (int): the threshold to use. Any immediates whose absolute values are larger than this threshold
            will be replaced with the `imm_str`
        include_negative (bool): if True, then any immediate that are too large and get replaced will have a negative
            sign added to the front of the replacement string if the immediates were negative
        imm_str (str): the string to replace immediate values with

    Returns:
        Union[Callable[..., str], str]: either a function that will handle thresholded immediate strings (if this 
            function was called correctly), or a handled thresholded immediate string
    """
    if not isinstance(threshold, int):
        raise TypeError('Threshold must be int, instead got `%s`: %s' % (type(threshold).__name__, threshold))

    def _threshold(self, state):
        # If we have the same value as the imm_str, then assume we've already been normalized
        if state.token == IMMEDIATE_VALUE_STR:
            return IMMEDIATE_VALUE_STR
        
        val = imm_to_int(state.token)
        return str(val) if abs(val) <= threshold else ('-' + imm_str) if val < 0 and include_negative else imm_str

    return _threshold


def replace_jump_destination(self, state):
    """Replaces the jump destination immediate with 'jmpdst' iff the jump destination is an immediate value, not a segment address

    Args:
        idx (int): the index in ``line`` of the 'jump' opcode
        line (List[TokenTuple]): a list of (token_type, token) tuples. the current assembly line

    Returns:
        int: integer index in line of last handled token
    """
    state.line[state.token_idx + 1] = (state.line[state.token_idx + 1][0], JUMP_DESTINATION_STR if state.line[state.token_idx + 1][0] == Tokens.IMMEDIATE else \
                                        state.line[state.token_idx + 1][1], state.line[state.token_idx + 1][2])
    return state.token_idx + 2


def replace_memory_expression(*args):
    """Builds a function that replaces memory expressions with the given replacement string
    
    This will return a function to be called as a part of a normalizer. This only takes one argument: the replacement string.
    If no arguments are passed, then the replacement string will default to 'memexpr'

    NOTE: This is meant to be a higher-order function. But, just in case the user forgets that (or is too lazy to add in
    two extra characters to call this function), if you pass multiple args then it will be assumed this is being called 
    as if it is the _repl_func() function below and will simply return the default result

    Args:
        args: args for this function. Ideally either empty to use default memory expression string, or a string to replace
            all memory expressions with.

    Returns:
        Union[Callable[..., None], None]: either a function that will handle memory expressions (if this function was 
            called correctly), or a handled memory expression
    """
    replacement = args[0] if len(args) == 1 else MEMORY_EXPRESSION_STR

    if not isinstance(replacement, str):
        raise TypeError('Replacement must be str, instead got `%s`: %s' % (type(replacement).__name__, replacement))

    def repl_mem(self, state):
        """Replace memory expressions with 'memexpr'"""
        # Using a brand new token so it's not confused with anything else. Keep track of the old value as well
        state.line[state.memory_start] = (Tokens.MEMORY_EXPRESSION, replacement, ' '.join([l[2] for l in state.line[state.memory_start + 1:]]))
        del state.line[state.memory_start + 1:]  # Delete the rest of the line after memory_start index

    return repl_mem if len(args) <= 1 else repl_mem(*args)


def replace_function_call_immediate(*args):
    """Builds a function that replaces function call immediate values with the given replacement string
    
    This will return a function to be called as a part of a normalizer. This only takes one argument: the replacement string.
    If no arguments are passed, then the replacement string will default to 'func'

    NOTE: This is meant to be a higher-order function. But, just in case the user forgets that (or is too lazy to add in
    two extra characters to call this function), if you pass multiple args then it will be assumed this is being called 
    as if it is the _repl_func() function below and will simply return the default result

    Args:
        args: args for this function. Ideally either empty to use default function call string, or a string to replace
            all function callsa with.

    Returns:
        Union[Callable[..., None], None]: either a function that will handle function calls (if this function was 
            called correctly), or a handled function call
    """
    replacement = args[0] if len(args) == 1 else FUNCTION_CALL_STR

    if not isinstance(replacement, str):
        raise TypeError('Replacement must be str, instead got `%s`: %s' % (type(replacement).__name__, replacement))

    def _repl_func(self, state):
        state.line[state.token_idx + 1] = (Tokens.IMMEDIATE, replacement, state.line[state.token_idx + 1][2])
        return state.token_idx + 2

    return _repl_func if len(args) <= 1 else _repl_func(*args)


def special_function_call(self, state, ret_only_call_type=False):
    """Handles special function calls
    
    Special external functions have their name kept. Recursive calls are replaced with 'self', other internal function 
    calls are replaced with 'internfunc', other external function calls are replaced with 'externfunc'. If a block has
    multiple function calls out, then it will be replaced with 'multifunc'.

    NOTE: This can all only happen if cfg and block information is passed. If it is not passed, then all function
    calls will be replaced with 'func'

    Args:
        idx (int): the index in ``line`` of the 'call' opcode
        line (List[TokenTuple]): a list of (token_type, token) tuples. the current assembly line
        special_functions (Set[str]): a set of string special function names.
        cfg (Union[CFG, MemCFG], optional): either a ``CFG`` or ``MemCFG`` object that these lines occur in. Used for 
            determining function calls to self, internal functions, and external functions. If not passed, then these 
            will not be used. Defaults to None.
        block (Union[CFGBasicBlock, int], optional): either a ``CFGBasicBlock`` or integer block_idx in a ``MemCFG`` 
            object. Used for determining function calls to self, internal functions, and external functions. If not 
            passed, then these will not be used. Defaults to None.
        ret_only_call_type (bool): if True, will return only the call type being used as a string. This is only for testing
            purposes and should likely not be used in normalization as this function already can handle the normalizing.
            This will return a string if it is not a special function call (for the appropriate function call type), or
            a tuple with one element for a special function call (the name of the special function). 

    Returns:
        int: integer index in line of last handled token
    """
    idx, line, block, cfg = state.token_idx, state.line, state.block, state.cfg

    # If cfg is None, then we cannot determine intern/extern/self function calls. Just return 'func' for function call
    if cfg is None:
        line[idx + 1] = (Tokens.IMMEDIATE, FUNCTION_CALL_STR, line[idx + 1][2])
        return idx + 2

    # Check that we even need to replace the value. We should only replace immediate values, so check to make sure
    #   the old_token starts with a digit
    if not ret_only_call_type and line[idx + 1][2][0] not in "0123456789":
        return idx + 2

    # Find the function call info in cfg
    if isinstance(cfg, bincfg.MemCFG):
        # Find the outgoing edges of this block. Normally the first edge is the function call edge, however we still need
        #   to check because there can be times where it only contains the normal return edge as the function call
        #   address couldn't be resolved to a known basic block
        func_call_block_inds, edge_types = cfg.get_block_edges_out(block, ret_edge_types=True)

        # Make sure the first value is in fact a function call
        if len(edge_types) > 0 and edge_types[0] == EdgeType.FUNCTION_CALL.value:

            # Check for a self call by comparing the two block's function indices
            func_call_block_idx = func_call_block_inds[0]
            self_call = cfg.block_func_idx[block] == cfg.block_func_idx[func_call_block_idx]
            extern_func_name = cfg.get_block_function_name(func_call_block_idx) if cfg.is_block_extern_function(func_call_block_idx) else None

            # Check for a multi-function call
            multi_call = len(edge_types[edge_types == EdgeType.FUNCTION_CALL.value]) > 1
        
        # Otherwise, we have no clue where the function call goes to, treat it as just some innerfunc
        else:
            self_call = False
            extern_func_name = None
            multi_call = False

    # Otherwise this is a plain CFG. Get the call address from the next immediate value, and look up its info in the CFG
    elif isinstance(cfg, bincfg.CFG):
        fc_out = block.get_sorted_edges(edge_types='function', direction='out')[0]

        if len(fc_out) > 1:
            multi_call = True
        else:
            multi_call = False

            # Need to check that the disassembler was able to figure out the call location
            if len(fc_out) == 0:
                self_call = False
                extern_func_name = None
            else:
                func = cfg.get_block(fc_out[0].to_block).parent_function
                self_call = func.address == block.parent_function.address
                extern_func_name = func.name if func.is_extern_function else None
    
    else:
        raise TypeError("Unknown cfg type for special_function_call normalization: %s" % repr(type(cfg)))

    # Determine what function call strings to return
    # A multi-function call
    if multi_call:
        if ret_only_call_type: return MULTI_FUNCTION_CALL_STR
        line[idx + 1] = (line[idx + 1][0], MULTI_FUNCTION_CALL_STR, line[idx + 1][2])

    # A recursive function call
    elif self_call:
        if ret_only_call_type: return RECURSIVE_FUNCTION_CALL_STR
        line[idx + 1] = (line[idx + 1][0], RECURSIVE_FUNCTION_CALL_STR, line[idx + 1][2])

    # An external function call
    elif extern_func_name is not None:
        extern_func_name = extern_func_name.split('@')[0]
        if extern_func_name in self.special_functions:
            if ret_only_call_type: return (extern_func_name,)
            line[idx + 1] = (line[idx + 1][0], '{' + extern_func_name + '}', line[idx + 1][2])
        else:
            if ret_only_call_type: return EXTERNAL_FUNCTION_CALL_STR
            line[idx + 1] = (line[idx + 1][0], EXTERNAL_FUNCTION_CALL_STR, line[idx + 1][2])

    # An internal function call
    else:
        if ret_only_call_type: return INTERNAL_FUNCTION_CALL_STR
        line[idx + 1] = (line[idx + 1][0], INTERNAL_FUNCTION_CALL_STR, line[idx + 1][2])

    return idx + 2



from ..base_tokenizer import RE_DISASSEMBLER_INFO, RE_SPACING, RE_NEWLINE, Tokens , BaseTokenizer, Architectures


# The delimiter lookahead. Allows for lookahead to make sure there is some delimiter, and we don't greedily grab 
#   registers/values/etc in case their names interfere with anything else.
# Can be a spacing, newline, or end of string
JAVA_RE_DELIMITER = r'{spacing}|{newline}|{dis_info}|$'.format(spacing=RE_SPACING, newline=RE_NEWLINE, dis_info=RE_DISASSEMBLER_INFO)

# Check for any opcode mnemonic. For now, we just get any and all characters that could correspond to an opcode, 
#   instruction prefix, or branch prediction. These will be sorted out later when parsing. Must be followed by delimiter
JAVA_RE_OPCODE = r'(?:[a-z][a-z0-9_]*)(?={delim})'.format(delim=JAVA_RE_DELIMITER)

# Check for the one 'prefix' opcode: wide
JAVA_RE_PREFIX = r'wide(?={delim})'.format(delim=JAVA_RE_DELIMITER)

# Match names to tokens, and define the order in which they should be matched
JAVA_DEFAULT_TOKENS = [
    (Tokens.INSTRUCTION_PREFIX, JAVA_RE_PREFIX),  # The 'wide' instruction prefix
    (Tokens.OPCODE, JAVA_RE_OPCODE),  # Opcode. Should go last since it's the most general matching. 
]
"""Default list of (token_type, regex) token tuples to match to"""


class JavaBaseTokenizer(BaseTokenizer):
    """A default class to tokenize java bytecode line input
    
    The tokenizer will tokenize essentially anything, so long as it fits known tokens.

    Known Tokens:

        * All of the default tokens from ``BaseTokenizer``
        * Instruction prefix: the 'wide' prefix
        * Opcode: any alpha-numeric + underscore substring

    Anything that does not fit one of the above tokens will be considered a 'token mismatch'


    Parameters
    ----------
    tokens: `Optional[List[Tuple[str, str]]]`
        the tokens to use. Should be a list of 2-tuples. Each tuple is a pair of (name, regex) where
        name is the string name of the token, and regex is a regular expression to find that token. These
        tuples should be ordered in the preferred order to search for tokens. If None, then this will default to 
        self.DEFAULT_TOKENS (which should be set when defining the class)
    token_handlers: `Optional[Dict[str, Callable[[Dict[str, Any]], Union[None, str]]]]`
        optional dictionary mapping token type strings to functions to handle those token types when tokenizing. This is
        intended to be used when you wish to add entirely new token types not present in `bincfg.normalization.base_tokenizer.Tokens`.
        If you wish to change the behavior of handling an already-present token type, just override that token handler function.
        These will override the default token handlers.
    insert_special_tokens: `bool`
        by default, some special tokens will be inserted at the front of `tokens` (see the 'special tokens' listed above).
        If you wish to stop this from happening, you can set `insert_special_tokens` to False
    case_sensitive: `bool`
        If True, then it is assumed that all regular expressions will exactly match case. If False, then it is assumed
        that all regular expressions only handle lowercase strings, and all incoming instructions will be converted to lowercase
    """
    DEFAULT_TOKENS = JAVA_DEFAULT_TOKENS

    ARCHITECTURE = Architectures.JAVA
    """The architecture this tokenizer works on"""
    

JAVA_DEFAULT_TOKENIZER = JavaBaseTokenizer()



import bincfg
import numpy as np
import sys
from .atomic_token_dict import AtomicTokenDict

# Types that can designate integers
INTEGER_TYPES = (int, np.integer)

# Whether or not we are in a typing version of python
IN_PYTHON_TYPING_VERSION = (sys.version_info.major, sys.version_info.minor) >= (3, 9)

# Only create types if in a python version that allows for type hinting nicely
if IN_PYTHON_TYPING_VERSION:
    from typing import Union, Protocol, Iterable, Optional, Any, Literal, NoReturn, Tuple, Mapping, Sequence, TextIO, Callable
    
    # Extra imports for type hints that don't necessarily have to be installed
    try:
        import torch
    except ImportError:
        pass
    try:
        import networkx
    except ImportError:
        pass
    try:
        from smda.common.SmdaReport import SmdaReport
        CFGInputDataType = Optional[Union[str, TextIO, Sequence[str], SmdaReport]]
    except ImportError:
        pass
    try:
        from typing_extensions import Self
    except ImportError:
        pass

    PlainAddress = Union[int, str]
    """Types that can be converted into an address by themselves, without having to look at any attributes"""

    class Addressable(Protocol):
        """Object that has a `.address` attribute which can be converted into a memory address"""
        address: 'PlainAddress'

    AddressLike = Union[PlainAddress, Addressable]
    """Objects that can be converted into a memory address, or that have a `.address` attribute which can"""


    class NormalizerType(Protocol):
        """Object that has a valid `.normalize()` function"""
        def normalize(self, *strings: 'str', cfg: 'Optional[bincfg.CFG]', 
                    block: 'Optional[bincfg.CFGBasicBlock]', newline_tup: 'Union[None, Tuple[str, str], object]', 
                    match_instruction_address: 'bool', **kwargs: Any) -> 'list[str]': ...


    TokenDictType = Union[dict[str, int], AtomicTokenDict]



"""
Classes/Methods involving edges in a ``CFG`` object
"""

import re
import bincfg
from ..utils import eq_obj
from enum import Enum
from ..utils.type_utils import *


# First value is None since the values start at 1
EDGE_TYPE_RES = [None, re.compile(r'(?:normal|plain|jump|branch)(?:[ -_]?edge)?'), re.compile(r'(?:call|fc|function|func)(?:[ -_]?call)?(?:[ -_]?edge)?')]


# Different edge types. Used when parsing a graphviz dot file
class EdgeType(Enum):
    """Enum for different edge types for ``CFGBasicBlock`` objects."""

    NORMAL = 1
    """a normal edge as a result of some branching/jumping instruction, or plain continuation to a next block
    
    (IE: an edge of control flow that does not involve calling a function)
    """

    FUNCTION_CALL = 2
    """an edge going from a basic block to another basic block in another function (or the same function). 
    
    The outgoing edge should always connect to a function entry block (IE: that block's ``.is_function_entry`` would be 
    True).
    """


def get_edge_type(edge_type: 'Union[EdgeType, str]') -> EdgeType:
    """Returns the edge type (instance of EdgeTypes enum class)

    Args:
        edge_type (Union[EdgeType, str]): can be either an EdgeTypes object, or a string. String values include:
            - 'normal': a EdgeTypes.NORMAL edge
            - 'function_call': a EdgeTypes.FUNCTION_CALL edge

    Raises:
        ValueError: for an unknown EdgeType string
        TypeError: for a bad `edge_type` type

    Returns:
        EdgeType: the given `edge_type` as a class from the EdgeType enum
    """
    if isinstance(edge_type, EdgeType):
        return edge_type
    elif isinstance(edge_type, int):
        for et in EdgeType:
            if et.value == edge_type:
                return et
        raise ValueError("Unknown EdgeType integer value: %d" % edge_type)
    elif isinstance(edge_type, str):
        lower_edge_type = edge_type.lower()
        for et in EdgeType:
            if EDGE_TYPE_RES[et.value].fullmatch(lower_edge_type):
                return et
        raise ValueError("Unknown EdgeType string: %s" % repr(edge_type))
    else:
        raise TypeError("Cannot get an EdgeType from object of type '%s'" % type(edge_type).__name__)


class CFGEdge:
    """A single immutable edge in a ``CFG`` object
    
    Parameters
    ----------
    from_block: `CFGBasicBlock`
        'from' ``CFGBasicBlock`` object
    to_block: `CFGBasicBlock`
        'to' ``CFGBasicBlock`` object
    edge_type: `Union[EdgeType, str]`
        the edge type. can be either an ``EdgeTypes`` object, or a string. String values include:
            - 'normal': a ``EdgeTypes.NORMAL`` edge
            - 'function_call': a ``EdgeTypes.FUNCTION_CALL`` edge
    """

    from_block: 'bincfg.CFGBasicBlock'
    """The from block of this directed edge"""
    to_block: 'bincfg.CFGBasicBlock'
    """The to block of this directed edge"""
    edge_type: 'EdgeType'
    """The type of this edge"""

    def __init__(self, from_block: 'bincfg.CFGBasicBlock', to_block: 'bincfg.CFGBasicBlock', edge_type: 'Union[EdgeType, str]'):
        if not isinstance(from_block, bincfg.CFGBasicBlock) or not isinstance(to_block, bincfg.CFGBasicBlock):
            raise TypeError("Both `from_block` and `to_block` inputs must be CFGBasicBlock's. Got: %s and %s"
                % (repr(type(from_block).__name__), repr(type(to_block).__name__)))
        
        self.from_block = from_block
        self.to_block = to_block
        self.edge_type = get_edge_type(edge_type)
        self._initialized: bool = True
    
    def __setattr__(self, name: 'str', value: 'Any') -> 'None':
        if hasattr(self, '_initialized') and self._initialized:
            raise TypeError("Cannot re-initialize immutable CFGEdge!")
        object.__setattr__(self, name, value)

    def __delattr__(self, name: 'str') -> 'NoReturn':
        raise TypeError("Cannot delete attributes on immutable CFGEdge!")
    
    @property
    def is_normal_edge(self) -> 'bool':
        """True if this is a 'normal' edge type, False otherwise"""
        return self.edge_type is EdgeType.NORMAL

    @property
    def is_function_call_edge(self) -> 'bool':
        """True if this is a 'function_call' edge type, False otherwise"""
        return self.edge_type is EdgeType.FUNCTION_CALL
    
    @property
    def is_branch(self) -> 'bool':
        """True if this edge is one of a branching instruction, False otherwise
        
        Specifically, returns True if this edge's `from_block` has exactly two outgoing edges, both of which are 'normal'
        edges. Sometimes, it is possible for blocks to have more than two 'normal' edges out (IE: jump tables), and 
        those are NOT considered branches and this method would return False
        """
        return len(self.from_block.edges_out) == 2 and all(e.edge_type is EdgeType.NORMAL for e in self.from_block.edges_out)
    
    def __str__(self) -> 'str':
        from_addr = ('0x%08x' % self.from_block.address if self.from_block.address is not None else 'NO_ADDRESS')
        to_addr = ('0x%08x' % self.to_block.address if self.to_block.address is not None else 'NO_ADDRESS')
        return "CFGEdge of type 'EdgeType.%s' from basic block at address %s to basic block at address %s" % \
            (self.edge_type.name, from_addr, to_addr)
    
    def __repr__(self) -> 'str':
        return self.__str__()
    
    def __eq__(self, other: 'Any') -> 'bool':
        """Equality for edges. Checks the to/from basic block addresses, and edge type"""
        return isinstance(other, CFGEdge) and \
            (self.from_block.address, self.to_block.address, self.edge_type.value) == (other.from_block.address, other.to_block.address, other.edge_type.value)
    
    def __hash__(self) -> 'int':
        return bincfg.hash_obj([self.from_block.address, self.to_block.address, self.edge_type.value], return_int=True)
    
    def __lt__(self, other: 'CFGEdge') -> 'bool':
        return (self.from_block.address, self.to_block.address, self.edge_type.value) < (other.from_block.address, other.to_block.address, other.edge_type.value)
    def __gt__(self, other: 'CFGEdge') -> 'bool':
        return (self.from_block.address, self.to_block.address, self.edge_type.value) > (other.from_block.address, other.to_block.address, other.edge_type.value)
    def __le__(self, other: 'CFGEdge') -> 'bool':
        return (self.from_block.address, self.to_block.address, self.edge_type.value) <= (other.from_block.address, other.to_block.address, other.edge_type.value)
    def __ge__(self, other: 'CFGEdge') -> 'bool':
        return (self.from_block.address, self.to_block.address, self.edge_type.value) >= (other.from_block.address, other.to_block.address, other.edge_type.value)



"""
Functions to parse cfg inputs into ``CFG`` objects.
"""

import re
import html
import sys
import bincfg
import copy
from .cfg_function import CFGFunction
from .cfg_basic_block import CFGBasicBlock
from .cfg_edge import CFGEdge, EdgeType
from ..utils import get_address, get_module


DIGRAPH_START_STRINGS = ['digraph', 'graph', 'node', 'edge']

# Regex matches
FUNC_STR_MATCH = re.compile(r'label=".*\\""')
FUNC_STR_MATCH_NO_NAME = re.compile(r'label="function 0x[0-9a-fA-F]*"')
ASM_LINE_MATCH = re.compile(r'label=<.*/>>')
FUNCTIONLESS_GV_BLOCK = re.compile(r'V_0x[0-9a-fA-F]* \[ .*')

# List of strings that are names for 'indeterminate' nodes
INDETERMINATE = ['indeterminate', 'nonexisting']

# Map a rose name to its in/out edge type ints
ROSE_EDGE_TYPES = {
    'call': EdgeType.FUNCTION_CALL,
    'cret': EdgeType.NORMAL,
    'cret\\nassumed': EdgeType.NORMAL,
    'cret\\\\nassumed': EdgeType.NORMAL,
    '': EdgeType.NORMAL,
    'other': EdgeType.NORMAL
}

# Regular expressions that denote an external function
EXTERN_FUNC_NAME_REGEXS = [re.compile(r'.*@plt'), re.compile(r'.*@.*[.]dll')]


def parse_cfg_data(cfg, data):
    """Parses the incoming cfg data. Infers type of data

    Args:
        cfg (CFG): the cfg to parse into
        data (Union[str, Sequence[str], TextIO, pd.DataFrame]): the data to parse, can be:

            - string: either string with newline characters that will be split on all newlines and treated as either a
              text or graphviz rose input, or a string with no newline characters that will be treated as a filename.
              Filenames will be opened as ghidra parquet files if they end with either '.pq' or '.parquet', and
              text/graphviz rose input otherwise
            - Sequence of string: will be treated as already-read-in text/graphviz rose input
            - open file object: will be read in using `.readlines`, then treated as text/graphviz rose input
            - pandas dataframe: will be parsed as ghidra parquet file

    Raises:
        ValueError: bad ``str`` filename, or an unknown file start string
        TypeError: bad ``data`` input type
        CFGParseError: if there is an error during CFG parsing (but data type was inferred correctly)
    """
    if isinstance(data, str):

        # Check for the empty string, and initialize empty
        if data == '':
            return
            
        # Check for single string to split on newlines
        if '\n' in data:
            data = [l.strip() for l in data.split('\n') if l.strip()]

        # Otherwise, assume it is a file
        else:
            cfg.metadata['filepath'] = data

            # Assume it is a text file
            try:
                with open(data, 'r') as f:
                    data = [l.strip() for l in f.readlines() if l.strip()]
            except:
                raise ValueError("Data was assumed to be a filename, but that file could not be opened/read!: %s" % repr(data))

    # Check for an open file
    elif hasattr(data, 'readlines') and callable(data.readlines):
        data = [l.strip() for l in data.readlines() if l.strip()]

    # Check for a copy constructor
    elif isinstance(data, bincfg.CFG):
        cfg.add_function(*[
            CFGFunction(address=func.address, name=func.name, is_extern_function=func._is_extern_function, metadata=copy.deepcopy(func.metadata), blocks=[
                CFGBasicBlock(
                    address=block.address,
                    edges_out=[(e.from_block.address, e.to_block.address, e.edge_type) for e in block.edges_out],
                    asm_lines=copy.deepcopy(block.asm_lines),
                    asm_memory_addresses=copy.deepcopy(block.asm_memory_addresses),
                    metadata=copy.deepcopy(block.metadata),
                ) for block in func.blocks
            ]) for func in data.functions_dict.values()
        ])

        # Also need to copy the metadata
        cfg.metadata = copy.deepcopy(data.metadata)

        # Also copy the normalizer
        cfg.normalizer = copy.deepcopy(data.normalizer)
        return

    # Check for a networkx to read in
    elif get_module('networkx', raise_err=False) and isinstance(data, (sys.modules['networkx'].DiGraph)):
        bincfg.CFG.from_networkx(data, cfg=cfg)
        return
    
    # Check for a smda report object from a disassembled file
    elif get_module('smda', raise_err=False) and isinstance(data, (sys.modules['smda'].common.SmdaReport.SmdaReport)):
        return parse_smda(data, cfg)

    # Otherwise, assume it is a sequence of string lines
    else:
        try:
            data = [l.strip() for l in data if l.strip()]
        except:
            raise TypeError("Could not parse CFG data from data of type: '%s'" % type(data).__name__)
    
    # If data is a list right now, assume we need to get the function from a list of lines
    if isinstance(data, list):
        func = _get_parse_func_from_lines(data)
    
    func(cfg, data)


def _get_parse_func_from_lines(lines):
    """Returns the function that should be used to parse this list of lines.

    Assumes all empty lines have already been stripped/removed

    Args:
        lines (Sequence[str]): the list of lines

    Returns:
        Callable[[CFG, Any], None]: the function to use to parse
    """

    # If lines is empty, return a function that does nothing
    if len(lines) == 0:
        return lambda *args, **kwargs: None
    
    # Otherwise check for different start lines
    else:
        if lines[0].startswith('digraph'):
            return parse_rose_gv
        elif lines[0].startswith('function'):
            return parse_rose_txt
        else:
            raise ValueError("Unknown file start string, could not infer file type!:\n%s\n..." % repr(lines[0][:100]))


##################
# Rose Text File #
##################


def parse_rose_txt(cfg, lines):
    """Reads input as a .txt file

    Args:
        cfg (CFG): an empty/loading CFG() object
        lines (str, Iterable[str], TextIO): the data to parse. Can be a string (which will be split on newlines to get each
            individual line), a list of string (each element will be considered one line), or an open file to call
            `.readlines()` on

    Raises:
        CFGParseError: when file does not fit expected format
    """
    if isinstance(lines, str):
        lines = lines.split('\n')
    elif hasattr(lines, 'readlines') and callable(lines, 'readlines'):
        lines = lines.readlines()

    try:
        # Clean up the lines a bit
        lines = [l.strip() for l in lines if l.strip()]
    except:
        raise TypeError("Could not parse rose txt input of type: '%s'" % type(lines).__name__)

    cfg.metadata['file_type'] = 'txt'
    
    # Make the dictionary of the current blocks
    curr_blocks = {}
    funcs = []
    
    # Go through lines finding each function
    curr_func_lines = [lines[0]]
    for line in lines[1:]:
        
        # Make the next function with the current list of lines
        if line.startswith('function 0x'):
            funcs.append(_parse_txt_function(cfg, curr_func_lines, curr_blocks))
            curr_func_lines = [line]
        else:
            curr_func_lines.append(line)
    
    funcs.append(_parse_txt_function(cfg, curr_func_lines, curr_blocks))
    cfg.add_function(*funcs)


def _parse_txt_function(cfg, func_lines, curr_blocks):
    """Parses the function lines from a rose txt file into a ``CFGFunction``, and returns the function

    Args:
        cfg (CFG): the ``CFG`` to which this function would belong
        func_lines (List[str]): list of string lines from file to parse for this function
        curr_blocks (Dict[int, CFGBasicBlock]): a dictionary mapping basic block addresses to ``CFGBasicBlock`` objects. 
            We need this to create new basic blocks on the fly in order to make ``CFGEdge``'s work properly

    Returns:
        CFGFunction: the cfg function
    """
    # Create the CFGFunction() object with its parent_cfg, name (while removing quotes from rose text), and is_extern_func
    _, address, *func_name_lines = func_lines[0].split(" ")
    name = ''.join(func_name_lines)[1:-1] if func_name_lines else None
    func = CFGFunction(parent_cfg=cfg, address=get_address(address), name=name, is_extern_function=_is_extern_func_name(name))

    # Build up every basic block
    curr_block_lines = [func_lines[1]]
    for line in func_lines[2:]:
        # Make the next block
        if line.startswith("B"):
            # Check if this is the first basic block and has the same starting address as the function
            _parse_txt_block(func, curr_block_lines, curr_blocks)
            curr_block_lines = [line]
        else:
            curr_block_lines.append(line)
    
    # Add in final block
    _parse_txt_block(func, curr_block_lines, curr_blocks)

    return func


def _parse_txt_block(func, block_lines, curr_blocks):
    """Parses the incoming block lines from a rose text file, and appends it to func's blocks

    Args:
        func (CFGFunction): the function to which this basic block belongs
        block_lines (List[str]): list of string block lines to build this block from
        curr_blocks (Dict[int, CFGBasicBlock]): a dictionary mapping basic block addresses to ``CFGBasicBlock`` objects. 
            We need this to create new basic blocks on the fly in order to make ``CFGEdge``'s work properly

    Raises:
        ValueError: on an unknown edge line
    """

    # Parse the block name and address, and check if it is a function entry point
    *_, block_address = block_lines[0].rpartition(" ")
    address = func.address if block_address[0] == 'p' else block_address[:-1] if block_address[-1] == ':' else block_address

    block = _create_basic_block(curr_blocks, address=get_address(address), parent_function=func)
    
    # If block is None, then this block already exists in another function, no need to recreate it
    if block is None:
        return

    for line in block_lines[1:]:

        # If this line is to tell us that this is a function return block "block is a function return/call"
        if line[0] == 'b':
            continue
        
        # This is an assembly line. Add the memory address and string line as a tuple
        # IMPORTANT: do this before the " edge " detection in case of string literals in rose <> info
        elif line.startswith("0x"):
            address, _, asm_line = line.partition(": ")
            block.asm_lines.append(asm_line.strip())
            block.asm_memory_addresses.append(get_address(address))
        
        # Currently just ignoring the 'also_owned_by' for now
        elif line[0] == 'a':
            #owned_by = line[23:].partition(" ")[0]
            #block.also_owned_by.add(int(owned_by, 16))
            pass
        
        # Otherwise this must be an edge line
        else:
            # Check for "function entry point", then lines using function names in quotes, then just normal address
            edge_addr = block.parent_function.address if line[-1] == 't' else \
                line.rpartition(' "')[0].rpartition(' ')[-1] if line[-1] == '"' else line.rpartition(' ')[-1]
            
            # edge_addr might already be an int from it's parent address
            if not isinstance(edge_addr, int):
                # Check for indeterminate/nonexistant edges. We ignore these, but check to see if this is a function return
                if edge_addr[0] != '0':
                    continue

                # Convert edge_addr to int
                edge_addr = int(edge_addr, 16)
            
            # Check for lines like "function call edge from/to", and "function return edge to" 
            if line[0] == 'f':
                if line[9] == 'c':
                    if line[19] == 't':
                        block.edges_out.add(CFGEdge(block, _create_basic_block(curr_blocks, edge_addr), 
                            edge_type=EdgeType.FUNCTION_CALL))
            
            # Check for "call return edge to" or "normal edge to"
            elif (line[0] == 'c' and line[17] == 't') or (line[0] == 'n' and line[12] == 't'):
                block.edges_out.add(CFGEdge(block, _create_basic_block(curr_blocks, edge_addr), edge_type=EdgeType.NORMAL))
            
            # Check to make sure this line is an 'edge from' line. Otherwise this is an unknown line, raise an error
            elif 'edge from' not in line:
                raise ValueError("Unknown edge line: %s" % repr(line))
    
    func.blocks.append(block)


def _create_basic_block(curr_blocks, address, **kwargs):
    """Checks if there is a basic block with the given address in `curr_blocks`, and if not, creates it. Returns the block

    If the block does exist, then any kwargs in ``kwargs`` will be updated in the CFGBasicBlock, unless that block already
    has a parent_func in which case None will be returned and no blocks will be updated

    Args:
        curr_blocks (Dict[int, CFGBasicBlock]): curr_blocks: a dictionary mapping basic block addresses to 
            ``CFGBasicBlock`` objects. We need this to create new basic blocks on the fly in order to make 
            ``CFGEdge``'s work properly
        address (int): the integer memory address of the new basic block
        kwargs (Any): extra kwargs to pass to ``CFGBasicBlock`` object creation, or to update an already existing 
            CFGBasicBlock

    Raises:
        ValueError: _description_

    Returns:
        CFGBasicBlock: _description_
    """
    if address not in curr_blocks:
        curr_blocks[address] = CFGBasicBlock(address=address, **kwargs)
    elif len(kwargs) == 0:
        return curr_blocks[address]
    elif curr_blocks[address].parent_function is not None:
        return None
    else:
        for k, v in kwargs.items():
            if k in ['parent_function', 'edges_in', 'edges_out', 'asm_lines', 'metadata', 'asm_memory_addresses']:
                setattr(curr_blocks[address], k, v)
            else:
                raise ValueError("Unknown basic block kwarg: %s" % repr(k))
    return curr_blocks[address]


######################
# Rose Graphviz File #
######################


def parse_rose_gv(cfg, lines):
    """Reads input as a graphviz file

    Args:
        cfg (CFG): an empty/loading CFG() object
        lines (str, Iterable[str], TextIO): the data to parse. Can be a string (which will be split on newlines to get each
            individual line), a list of string (each element will be considered one line), or an open file to call
            `.readlines()` on

    Raises:
        CFGParseError: when the file cannot be parsed correctly
    """
    if isinstance(lines, str):
        lines = lines.split('\n')
    elif hasattr(lines, 'readlines') and callable(lines, 'readlines'):
        lines = lines.readlines()

    try:
        # Clean up the lines a bit
        lines = [l.strip() for l in lines if l.strip()]
    except:
        raise TypeError("Could not parse rose graphviz input of type: '%s'" % type(lines).__name__)

    cfg.metadata['file_type'] = 'gv'

    subgraphs = []
    edges = {}
    curr_blocks = {}
    
    # Keeping track of states
    in_subgraph = False
    eof = False

    for line in lines:
        # Check for empty string, for beginning digraph strings to ignore, and indeterminate/nonexisting nodes
        if line == '' or any(line.startswith(s) for s in DIGRAPH_START_STRINGS + INDETERMINATE):
            continue
        
        # Check for subgraph cluster
        elif line[0] == 's':
            in_subgraph = True

            # Get the function string
            func_str_matches = FUNC_STR_MATCH.findall(line)

            # Check for functions with no name
            if len(func_str_matches) == 0:
                func_str_matches = FUNC_STR_MATCH_NO_NAME.findall(line)
            
            # Otherwise continue normally
            if len(func_str_matches) != 1:
                raise CFGParseError("Could not parse function string from: %s\n Found matches: %s" % (repr(line), func_str_matches))

            # Add a new subgraph to the list (getting the [7:-1] works in both named and unnamed cases)
            # The func_str should be something like 'function [MEMORY_ADDRESS] "[FUNCTION_NAME]"' or 'function [MEMORY_ADDRESS]'
            func_str = func_str_matches[0][7:-1]
            _, func_address, *func_name = func_str.split(' ')
            func_name = ' '.join(func_name)[2:-2] if func_name else None

            subgraphs.append((func_name, int(func_address, 0), []))
        
        # Check for end of subgraph cluster/eof
        elif line[0] == '}':
            # Check to make sure there is only one eof '}' line
            if not in_subgraph:
                if not eof:
                    eof = True
                else:
                    raise CFGParseError("Found multiple lines starting with '}' that did not end subgraphs")

            in_subgraph = False
        
        # Check for nodes/node edges
        elif line[0] == 'V':
            
            # Handle subgraph node, or if it is a block with no parent function
            if in_subgraph or FUNCTIONLESS_GV_BLOCK.fullmatch(line) is not None:

                # Get the node address
                address, _, rest = line.partition(" [ ")
                address = int(address[2:], 0)
                
                # Get the asm line string
                asm_line_match = ASM_LINE_MATCH.findall(rest)

                # Need to leave the first and last <>
                asm_line = asm_line_match[0][6:] if len(asm_line_match) > 0 else ''

                # The tuple for this current node
                node_tup = (address, asm_line)

                # Add this node to our current subgraph if we are in one
                if in_subgraph:
                    subgraphs[-1][2].append(node_tup)
                
                # Otherwise, we are parsing a functionless basic block, create a dummy function to wrap it
                else:
                    subgraphs.append(("__DUMMY_FUNCTION_AT_0x%x__" % address, address, [node_tup]))

            # Handle edge
            else:
                # Get the source and destination names
                source, rest = [a.strip() for a in line.split('->')]
                dest, rest = [a.strip() for a in rest.split(' [ ')]

                # Get the label name by splitting on quotes and getting first index, checking for empty string as well
                label = "" if 'label=""' in rest else rest.split('"')[1]

                # Don't deal with indeterminate edges, unless they are a function return, then send that info
                if dest in INDETERMINATE:
                    continue

                source, dest = int(source[2:], 0), int(dest[2:], 0)

                # Add the edge into the dictionary for the outgoing edge
                edges.setdefault(source, []).append((ROSE_EDGE_TYPES[label], dest))

        # Otherwise, raise error
        else:
            raise CFGParseError("Unknown line: '%s'" % line)
    
    funcs = [_parse_gv_function(cfg, name, address, nodes, edges, curr_blocks) for name, address, nodes in subgraphs]
    cfg.add_function(*funcs)


def _parse_gv_function(cfg, name, address, nodes, edges, curr_blocks):
    """Parses the func_info as a graphviz dot file, returns the function

    Args:
        cfg (CFG): the ``CFG`` to which this function belongs
        name (Union[str, None]): the function name, or None if it doesn't have one
        address (int): the integer address of this function
        nodes (Iterable[Tuple[int, str]]): an iterable of nodes in this subgraph. Each 'node' should be a tuple of 
            (node_address: int, node_asm_lines: str), with the 'node_asm_lines' being the unprocessed string from the 
            graphviz file
        edges (Dict[int, List[Tuple[EdgeType, int]]]): a dictionary of all edges in the cfg. Each key should be a 'from'
            basic block integer address, and values are tuples of outgoing edge information for the block with that 
            address. Each edge information is a tuple of (edge_type: EdgeType, to_address: int)
        curr_blocks (Dict[int, CFGBasicBlock]): a dictionary mapping basic block addresses to ``CFGBasicBlock`` objects. 
            We need this to create new basic blocks on the fly in order to make ``CFGEdge``'s work properly

    Returns:
        CFGFunction: the cfg function
    """
    func = CFGFunction(parent_cfg=cfg, address=get_address(address), name=name, is_extern_function=_is_extern_func_name(name))
    for address, asm_lines in nodes:
        _parse_gv_block(func, address, asm_lines, edges.get(address, []), curr_blocks)
    
    return func


def _parse_gv_block(func, address, asm_lines, node_edges, curr_blocks):
    """Parses the incoming block info assuming it is from a graphviz dot file, and appends it to func's blocks

    Args:
        func (CFGFunction): the ``CFGFunction`` this block belongs to
        address (int): integer memory address of the node
        asm_lines (str): the UNPARSED asm lines from the raw gv dot file
        node_edges (Iterable[Tuple[EdgeType, int]]): an iterable of information for all outgoing edges for this block. 
            Each element should be a tuple of (edge_type: EdgeType, to_address: int)
        curr_blocks (Dict[int, CFGBasicBlock]): a dictionary mapping basic block addresses to ``CFGBasicBlock`` objects. 
            We need this to create new basic blocks on the fly in order to make ``CFGEdge``'s work properly
    """
    # Get the CFGBasicBlock with this address
    asm_stuff = {k: v for k, v in zip(['asm_lines', 'asm_memory_addresses'], get_asm_from_node_label(asm_lines))}
    block = _create_basic_block(curr_blocks, address, parent_function=func, **asm_stuff)

    # Parse out the edges
    for edge_type, address in node_edges:
        block.edges_out.add(CFGEdge(block, _create_basic_block(curr_blocks, address), edge_type))
    
    func.blocks.append(block)


GV_SPLIT = re.compile(r'<br [^>]*/>')
def get_asm_from_node_label(label):
    """Converts a node's label into a list of assembly lines at that basic block.

    Args:
        label (str): the unparsed string label

    Returns:
        Tuple[List[str], List[int]]: tuple of 2 lists: (asm_lines, asm_memory_addresses)
    """
    if label == '' or label is None:
        return []

    # Remove the first and last <>, replace all "??" with empty string, and html-unescape the ampersand encoded things
    ret = [('0x' + html.unescape(l.replace("??", ""))) for l in GV_SPLIT.split(label[1:-1]) if l != ""]

    # Split on spaces and get the first one to get the memory address, the rest are joined to be the instruction
    lines = [line.strip() for r in ret for addr, _, line in [r.replace('\t', '').partition(' ')]]
    addrs = [int(addr, 0) for r in ret for addr, _, line in [r.replace('\t', '').partition(' ')]]
    return lines, addrs


def _is_extern_func_name(name):
    """Returns True if name is an external function name, False otherwise
    
    Args:
        name (Union[str, None]): the name
    
    Returns:
        bool: True if name is an external function name, False otherwise
    """
    return name is not None and any(s.fullmatch(name) is not None for s in EXTERN_FUNC_NAME_REGEXS)


class CFGParseError(Exception):
    pass
